#!/usr/bin/python
# -*- coding: utf-8 -*-

from cryptos import *
from wicc.language import Language


def random_words(language: Language=Language.English):
    return entropy_to_words(os.urandom(16), language_words_list(language=language))


def language_words_list(language: Language):
    if language.name == Language.French.name:
        from wicc.cryptos.words_fr import words_list
        return words_list
    elif language.name == Language.Spanish.name:
        from wicc.cryptos.words_es import words_list
        return words_list
    elif language.name == Language.Italian.name:
        from wicc.cryptos.words_it import words_list
        return words_list
    elif language.name == Language.Korean.name:
        from wicc.cryptos.words_ko import words_list
        return words_list
    elif language.name == Language.Japanese.name:
        from wicc.cryptos.words_jp import words_list
        return words_list
    elif language.name == Language.Simplified_Chinese.name:
        from wicc.cryptos.words_zh_hans import words_list
        return words_list
    else:
        from wicc.cryptos.words_en import words_list
        return words_list


def to_en_mnemonic_string(language: Language, mnemonic):
    if language == Language.English:
        return mnemonic
    target_words = []
    en_words_list = language_words_list(language=Language.English)
    words_list = language_words_list(language=language)
    for word in mnemonic.split(" "):
        for (index, word_in) in enumerate(words_list):
            if word == word_in:
                target_words.append(en_words_list[index])
    print(target_words)
    return " ".join(target_words)




from ..explorers import blockchain
from .base import BaseCoin


class Bitcoin(BaseCoin):
    coin_symbol = "BTC"
    display_name = "Bitcoin"
    segwit_supported = True
    explorer = blockchain
    magicbyte = 0
    script_magicbyte = 5
    segwit_hrp = "bc"
    client_kwargs = {
        'server_file': 'bitcoin.json',
    }

    testnet_overrides = {
        'display_name': "Bitcoin Testnet",
        'coin_symbol': "BTCTEST",
        'magicbyte': 111,
        'script_magicbyte': 196,
        'segwit_hrp': 'tb',
        'hd_path': 1,
        'wif_prefix': 0xef,
        'client_kwargs': {
            'server_file': 'bitcoin_testnet.json',
        },
        'xprv_headers': {
            'p2pkh': 0x04358394,
            'p2wpkh-p2sh': 0x044a4e28,
            'p2wsh-p2sh': 0x295b005,
            'p2wpkh': 0x04358394,
            'p2wsh': 0x2aa7a99
        },
        'xpub_headers': {
            'p2pkh': 0x043587cf,
            'p2wpkh-p2sh': 0x044a5262,
            'p2wsh-p2sh': 0x295b43f,
            'p2wpkh': 0x043587cf,
            'p2wsh': 0x2aa7ed3
        },
    }


import requests
import demjson
import json

from wicc.wiccrpc.box.Config import Config


class RpcManager(object):
    @classmethod
    def request(cls, parameters):
        parameters["jsonrpc"] = "2.0"
        parameters["id"] = "curltext"
        data = json.dumps(parameters)
        resp = requests.post(Config.url(), headers=Config.header(), data=data)
        return demjson.decode(resp.text)





# coding=utf-8
from wicc.wiccrpc.box.RpcManager import RpcManager


class Vote(object):
    """
    钱包即账号的集合，可以管理多个账号
    """
    @classmethod
    def vote(cls, sender, receiver, ticket_amount, fee=10000):
        """
        投票，手续费默认为:10000sawi
        :param sender: 发送者地址
        :param receiver: 投票地址
        :param amount: 发送金额(单位sawi)
        :param fee: 交易费用(单位sawi),最少10000sawi
        :return: 交易Hash
        """
        data = {
            "method": "votedelegatetx",
            "params": [sender, [{"delegate": receiver, "votes": ticket_amount*100000000}], fee]
        }
        return RpcManager.request(data)





# Copyright (c) 2016-2017, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Classes for acting as a peer over a transport and speaking the JSON
RPC versions 1.0 and 2.0.
JSONSessionBase can use an arbitrary transport.
JSONSession integrates asyncio.Protocol to provide the transport.
'''

import asyncio
import collections
import inspect
import json
import numbers
import time
import traceback

from . import util


class RPCError(Exception):
    '''RPC handlers raise this error.'''
    def __init__(self, msg, code=-1, **kw_args):
        super().__init__(**kw_args)
        self.msg = msg
        self.code = code


class JSONRPC(object):
    '''Base class of JSON RPC versions.'''

    # See http://www.jsonrpc.org/specification
    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_ARGS = -32602
    INTERNAL_ERROR = -32603

    # Codes for this library
    INVALID_RESPONSE = -100
    ERROR_CODE_UNAVAILABLE = -101
    REQUEST_TIMEOUT = -102
    FATAL_ERROR = -103

    ID_TYPES = (type(None), str, numbers.Number)
    HAS_BATCHES = False

    @classmethod
    def canonical_error(cls, error):
        '''Convert an error to a JSON RPC 2.0 error.
        Handlers then only have a single form of error to deal with.
        '''
        if isinstance(error, int):
            error = {'code': error}
        elif isinstance(error, str):
            error = {'message': error}
        elif not isinstance(error, dict):
            error = {'data': error}
        error['code'] = error.get('code', JSONRPC.ERROR_CODE_UNAVAILABLE)
        error['message'] = error.get('message', 'error message unavailable')
        return error

    @classmethod
    def timeout_error(cls):
        return {'message': 'request timed out',
                'code': JSONRPC.REQUEST_TIMEOUT}


class JSONRPCv1(JSONRPC):
    '''JSON RPC version 1.0.'''

    @classmethod
    def request_payload(cls, id_, method, params=None):
        '''JSON v1 request payload.  Params is mandatory.'''
        return {'method': method, 'params': params or [], 'id': id_}

    @classmethod
    def notification_payload(cls, method, params=None):
        '''JSON v1  notification payload.  Params and id are mandatory.'''
        return {'method': method, 'params': params or [], 'id': None}

    @classmethod
    def response_payload(cls, result, id_):
        '''JSON v1 response payload.  error is present and None.'''
        return {'id': id_, 'result': result, 'error': None}

    @classmethod
    def error_payload(cls, message, code, id_):
        '''JSON v1 error payload.  result is present and None.'''
        return {'id': id_, 'result': None,
                'error': {'message': message, 'code': code}}

    @classmethod
    def handle_response(cls, handler, payload):
        '''JSON v1 response handler.  Both 'error' and 'result'
        should exist with exactly one being None.
        Unfortunately many 1.0 clients behave like 2.0, and just send
        one or the other.
        '''
        error = payload.get('error')
        if error is None:
            handler(payload.get('result'), None)
        else:
            handler(None, cls.canonical_error(error))

    @classmethod
    def is_request(cls, payload):
        '''Returns True if the payload (which has a method) is a request.
        False means it is a notification.'''
        return payload.get('id') is not None


class JSONRPCv2(JSONRPC):
    '''JSON RPC version 2.0.'''

    HAS_BATCHES = True

    @classmethod
    def request_payload(cls, id_, method, params=None):
        '''JSON v2 request payload.  Params is optional.'''
        payload = {'jsonrpc': '2.0', 'method': method, 'id': id_}
        if params:
            payload['params'] = params
        return payload

    @classmethod
    def notification_payload(cls, method, params=None):
        '''JSON v2  notification payload.  There must be no id.'''
        payload = {'jsonrpc': '2.0', 'method': method}
        if params:
            payload['params'] = params
        return payload

    @classmethod
    def response_payload(cls, result, id_):
        '''JSON v2 response payload.  error is not present.'''
        return {'jsonrpc': '2.0', 'id': id_, 'result': result}

    @classmethod
    def error_payload(cls, message, code, id_):
        '''JSON v2 error payload.  result is not present.'''
        return {'jsonrpc': '2.0', 'id': id_,
                'error': {'message': message, 'code': code}}

    @classmethod
    def handle_response(cls, handler, payload):
        '''JSON v2 response handler.  Exactly one of 'error' and 'result'
        must exist.  Errors must have 'code' and 'message' members.
        '''
        if 'error' in payload:
            handler(None, cls.canonical_error(payload['error']))
        elif 'result' in payload:
            handler(payload['id'], payload['result'], None)
        else:
            error = {'message': 'no error or result returned',
                     'code': JSONRPC.INVALID_RESPONSE}
            handler(None, cls.canonical_error(error))

    @classmethod
    def batch_size(cls, parts):
        '''Return the size of a JSON batch from its parts.'''
        return sum(len(part) for part in parts) + 2 * len(parts)

    @classmethod
    def batch_bytes(cls, parts):
        '''Return the bytes of a JSON batch from its parts.'''
        if parts:
            return b'[' + b', '.join(parts) + b']'
        return b''

    @classmethod
    def is_request(cls, payload):
        '''Returns True if the payload (which has a method) is a request.
        False means it is a notification.'''
        return 'id' in payload


class JSONRPCCompat(JSONRPC):
    '''Intended to be used until receiving a response from the peer, at
    which point detect_version should be used to choose which version
    to use.
    Sends requests compatible with v1 and v2.  Errors cannot be
    compatible so v2 errors are sent.
    Does not send responses or notifications, nor handle responses.
    '''
    @classmethod
    def request_payload(cls, id_, method, params=None):
        '''JSON v2 request payload but with params present.'''
        return {'jsonrpc': '2.0', 'id': id_,
                'method': method, 'params': params or []}

    @classmethod
    def error_payload(cls, message, code, id_):
        '''JSON v2 error payload.  result is not present.'''
        return {'jsonrpc': '2.0', 'id': id_,
                'error': {'message': message, 'code': code}}

    @classmethod
    def detect_version(cls, payload):
        '''Return a best guess at a version compatible with the received
        payload.
        Return None if one cannot be determined.
        '''
        def item_version(item):
            if isinstance(item, dict):
                version = item.get('jsonrpc')
                if version is None:
                    return JSONRPCv1
                if version == '2.0':
                    return JSONRPCv2
            return None

        if isinstance(payload, list) and payload:
            version = item_version(payload[0])
            # If a batch return at least JSONRPCv2
            if version in (JSONRPCv1, None):
                version = JSONRPCv2
        else:
            version = item_version(payload)

        return version


class JSONSessionBase(util.LoggedClass):
    '''Acts as the application layer session, communicating via JSON RPC
    over an underlying transport.
    Processes incoming and sends outgoing requests, notifications and
    responses.  Incoming messages are queued.  When the queue goes
    from empty
    '''
    _next_session_id = 0
    _pending_reqs = {}    # Outgoing requests waiting for a response

    @classmethod
    def next_session_id(cls):
        '''Return the next unique session ID.'''
        session_id = cls._next_session_id
        cls._next_session_id += 1
        return session_id

    def _pending_request_keys(self):
        '''Return a generator of pending request keys for this session.'''
        return [key for key in self._pending_reqs if key[0] is self]

    def has_pending_requests(self):
        '''Return True if this session has pending requests.'''
        return bool(self._pending_request_keys())

    def pop_response_handler(self, msg_id):
        '''Return the response handler for the given message ID.'''
        return self._pending_reqs.pop((self, msg_id), (None, None))[0]

    def timeout_session(self):
        '''Trigger timeouts for all of the session's pending requests.'''
        self._timeout_requests(self._pending_request_keys())

    @classmethod
    def timeout_check(cls):
        '''Trigger timeouts where necessary for all pending requests.'''
        now = time.time()
        keys = [key for key, value in cls._pending_reqs.items()
                if value[1] < now]
        cls._timeout_requests(keys)

    @classmethod
    def _timeout_requests(cls, keys):
        '''Trigger timeouts for the given lookup keys.'''
        values = [cls._pending_reqs.pop(key) for key in keys]
        handlers = [handler for handler, timeout in values]
        timeout_error = JSONRPC.timeout_error()
        for handler in handlers:
            handler(None, timeout_error)

    def __init__(self, version=JSONRPCCompat):
        super().__init__()

        # Parts of an incomplete JSON line.  We buffer them until
        # getting a newline.
        self.parts = []
        self.version = version
        self.log_me = False
        self.session_id = None
        # Count of incoming complete JSON requests and the time of the
        # last one.  A batch counts as just one here.
        self.last_recv = time.time()
        self.send_count = 0
        self.send_size = 0
        self.recv_size = 0
        self.recv_count = 0
        self.error_count = 0
        self.pause = False
        # Handling of incoming items
        self.items = collections.deque()
        self.items_events ={}
        self.items_event = asyncio.Event()
        self.batch_results = []
        # Handling of outgoing requests
        self.next_request_id = 0
        # If buffered incoming data exceeds this the connection is closed
        self.max_buffer_size = 1000000
        self.max_send = 50000
        self.close_after_send = False

    def pause_writing(self):
        '''Transport calls when the send buffer is full.'''
        self.log_info('pausing processing whilst socket drains')
        self.pause = True

    def resume_writing(self):
        '''Transport calls when the send buffer has room.'''
        self.log_info('resuming processing')
        self.pause = False

    def is_oversized(self, length, id_):
        '''Return an error payload if the given outgoing message size is too
        large, or False if not.
        '''
        if self.max_send and length > max(1000, self.max_send):
            msg = 'response too large (at least {:d} bytes)'.format(length)
            return self.error_bytes(msg, JSONRPC.INVALID_REQUEST, id_)
        return False

    def send_binary(self, binary):
        '''Pass the bytes through to the transport.
        Close the connection if close_after_send is set.
        '''
        if self.is_closing():
            return
        self.using_bandwidth(len(binary))
        self.send_count += 1
        self.send_size += len(binary)
        self.send_bytes(binary)
        if self.close_after_send:
            self.close_connection()

    def payload_id(self, payload):
        '''Extract and return the ID from the payload.
        Returns None if it is missing or invalid.'''
        try:
            return self.check_payload_id(payload)
        except RPCError:
            return None

    def check_payload_id(self, payload):
        '''Extract and return the ID from the payload.
        Raises an RPCError if it is missing or invalid.'''
        if 'id' not in payload:
            raise RPCError('missing id', JSONRPC.INVALID_REQUEST)

        id_ = payload['id']
        if not isinstance(id_, self.version.ID_TYPES):
            raise RPCError('invalid id type {}'.format(type(id_)),
                           JSONRPC.INVALID_REQUEST)
        return id_

    def request_bytes(self, id_, method, params=None):
        '''Return the bytes of a JSON request.'''
        payload = self.version.request_payload(id_, method, params)
        return self.encode_payload(payload)

    def notification_bytes(self, method, params=None):
        payload = self.version.notification_payload(method, params)
        return self.encode_payload(payload)

    def response_bytes(self, result, id_):
        '''Return the bytes of a JSON response.'''
        return self.encode_payload(self.version.response_payload(result, id_))

    def error_bytes(self, message, code, id_=None):
        '''Return the bytes of a JSON error.
        Flag the connection to close on a fatal error or too many errors.'''
        version = self.version
        self.error_count += 1
        if not self.close_after_send:
            fatal_log = None
            if code in (version.PARSE_ERROR, version.INVALID_REQUEST,
                        version.FATAL_ERROR):
                fatal_log = message
            elif self.error_count >= 10:
                fatal_log = 'too many errors, last: {}'.format(message)
            if fatal_log:
                self.log_info(fatal_log)
                self.close_after_send = True
        return self.encode_payload(self.version.error_payload
                                   (message, code, id_))

    def encode_payload(self, payload):
        '''Encode a Python object as binary bytes.'''
        assert isinstance(payload, dict)

        id_ = payload.get('id')
        try:
            binary = json.dumps(payload).encode()
        except TypeError:
            msg = 'JSON encoding failure: {}'.format(payload)
            self.log_error(msg)
            binary = self.error_bytes(msg, JSONRPC.INTERNAL_ERROR, id_)

        error_bytes = self.is_oversized(len(binary), id_)
        return error_bytes or binary

    def decode_message(self, payload):
        '''Decode a binary message and pass it on to process_single_item or
        process_batch as appropriate.
        Messages that cannot be decoded are logged and dropped.
        '''
        try:
            payload = payload.decode()
        except UnicodeDecodeError as e:
            msg = 'cannot decode message: {}'.format(e)
            self.send_error(msg, JSONRPC.PARSE_ERROR)
            return

        try:
            payload = json.loads(payload)
        except json.JSONDecodeError as e:
            msg = 'cannot decode JSON: {}'.format(e)
            self.send_error(msg, JSONRPC.PARSE_ERROR)
            return

        if self.version is JSONRPCCompat:
            # Attempt to detect peer's JSON RPC version
            version = self.version.detect_version(payload)
            if not version:
                version = JSONRPCv2
                self.log_info('unable to detect JSON RPC version, using 2.0')
            self.version = version

        # Batches must have at least one object.
        if payload == [] and self.version.HAS_BATCHES:
            self.send_error('empty batch', JSONRPC.INVALID_REQUEST)
            return

        self.items.append(payload)
        self.items_events[payload['id']].set()

    async def process_batch(self, batch, count):
        '''Processes count items from the batch according to the JSON 2.0
        spec.
        If any remain, puts what is left of the batch back in the deque
        and returns None.  Otherwise returns the binary batch result.'''
        results = self.batch_results
        self.batch_results = []

        for n in range(count):
            item = batch.pop()
            result = await self.process_single_item(item)
            if result:
                results.append(result)

        if not batch:
            return self.version.batch_bytes(results)

        error_bytes = self.is_oversized(self.batch_size(results), None)
        if error_bytes:
            return error_bytes

        self.items.appendleft(item)
        self.batch_results = results
        return None

    async def process_single_item(self, payload):
        '''Handle a single JSON request, notification or response.
        If it is a request, return the binary response, oterhwise None.'''
        if self.log_me:
            self.log_info('processing {}'.format(payload))

        if not isinstance(payload, dict):
            return self.error_bytes('request must be a dictionary',
                                    JSONRPC.INVALID_REQUEST)

        try:
            # Requests and notifications must have a method.
            if 'method' in payload:
                if self.version.is_request(payload):
                    return await self.process_single_request(payload)
                else:
                    await self.process_single_notification(payload)
            else:
                self.process_single_response(payload)

            return None
        except asyncio.CancelledError:
            raise
        except Exception:
            self.log_error(traceback.format_exc())
            return self.error_bytes('internal error processing request',
                                    JSONRPC.INTERNAL_ERROR,
                                    self.payload_id(payload))

    async def process_single_request(self, payload):
        '''Handle a single JSON request and return the binary response.'''
        try:
            result = await self.handle_payload(payload, self.request_handler)
            return self.response_bytes(result, payload['id'])
        except RPCError as e:
            return self.error_bytes(e.msg, e.code, self.payload_id(payload))
        except asyncio.CancelledError:
            raise
        except Exception:
            self.log_error(traceback.format_exc())
            return self.error_bytes('internal error processing request',
                                    JSONRPC.INTERNAL_ERROR,
                                    self.payload_id(payload))

    async def process_single_notification(self, payload):
        '''Handle a single JSON notification.'''
        try:
            await self.handle_payload(payload, self.notification_handler)
        except RPCError:
            pass
        except Exception:
            self.log_error(traceback.format_exc())

    def process_single_response(self, payload):
        '''Handle a single JSON response.'''
        try:
            id_ = self.check_payload_id(payload)
            handler = self.pop_response_handler(id_)
            if handler:
                self.version.handle_response(handler, payload)
            else:
                self.log_info('response for unsent id {}'.format(id_),
                              throttle=True)
        except RPCError:
            pass
        except Exception:
            self.log_error(traceback.format_exc())

    async def handle_payload(self, payload, get_handler):
        '''Handle a request or notification payload given the handlers.'''
        # An argument is the value passed to a function parameter...
        args = payload.get('params', [])
        method = payload.get('method')

        if not isinstance(method, str):
            raise RPCError("invalid method type {}".format(type(method)),
                           JSONRPC.INVALID_REQUEST)

        handler = get_handler(method)
        if not handler:
            raise RPCError("unknown method: '{}'".format(method),
                           JSONRPC.METHOD_NOT_FOUND)

        if not isinstance(args, (list, dict)):
            raise RPCError('arguments should be an array or dictionary',
                           JSONRPC.INVALID_REQUEST)

        params = inspect.signature(handler).parameters
        names = list(params)
        min_args = sum(p.default is p.empty for p in params.values())

        if len(args) < min_args:
            raise RPCError('too few arguments to {}: expected {:d} got {:d}'
                           .format(method, min_args, len(args)),
                           JSONRPC.INVALID_ARGS)

        if len(args) > len(params):
            raise RPCError('too many arguments to {}: expected {:d} got {:d}'
                           .format(method, len(params), len(args)),
                           JSONRPC.INVALID_ARGS)

        if isinstance(args, list):
            kw_args = {name: arg for name, arg in zip(names, args)}
        else:
            kw_args = args
            bad_names = ['<{}>'.format(name) for name in args
                         if name not in names]
            if bad_names:
                raise RPCError('invalid parameter names: {}'
                               .format(', '.join(bad_names)))

        if inspect.iscoroutinefunction(handler):
            return await handler(**kw_args)
        else:
            return handler(**kw_args)

    # ---- External Interface ----

    async def process_pending_items(self, limit=8):
        '''Processes up to LIMIT pending items asynchronously.'''
        while limit > 0 and self.items:
            item = self.items.popleft()

            if isinstance(item, list) and self.version.HAS_BATCHES:
                count = min(limit, len(item))
                binary = await self.process_batch(item, count)
                limit -= count
            else:
                binary = await self.process_single_item(item)
                limit -= 1

            if binary:
                self.send_binary(binary)

    def count_pending_items(self):
        '''Counts the number of pending items.'''
        return sum(len(item) if isinstance(item, list) else 1
                   for item in self.items)

    def connection_made(self):
        '''Call when an incoming client connection is established.'''
        self.session_id = self.next_session_id()
        self.log_prefix = '[{:d}] '.format(self.session_id)

    def data_received(self, data):
        '''Underlying transport calls this when new data comes in.
        Look for newline separators terminating full requests.
        '''
        if self.is_closing():
            return
        self.using_bandwidth(len(data))
        self.recv_size += len(data)

        # Close abusive connections where buffered data exceeds limit
        buffer_size = len(data) + sum(len(part) for part in self.parts)
        if buffer_size > self.max_buffer_size:
            self.log_error('read buffer of {:,d} bytes over {:,d} byte limit'
                           .format(buffer_size, self.max_buffer_size))
            self.close_connection()
            return

        while True:
            npos = data.find(ord('\n'))
            if npos == -1:
                self.parts.append(data)
                break
            tail, data = data[:npos], data[npos + 1:]
            parts, self.parts = self.parts, []
            parts.append(tail)
            self.recv_count += 1
            self.last_recv = time.time()
            self.decode_message(b''.join(parts))

    def send_error(self, message, code, id_=None):
        '''Send a JSON error.'''
        self.send_binary(self.error_bytes(message, code, id_))

    def send_request(self, handler, method, params=None, timeout=30):
        '''Sends a request and arranges for handler to be called with the
        response when it comes in.
        A call to request_timeout_check() will result in pending
        responses that have been waiting more than timeout seconds to
        call the handler with a REQUEST_TIMEOUT error.
        '''
        id_ = self.next_request_id
        self.next_request_id += 1
        self.send_binary(self.request_bytes(id_, method, params))
        self.items_events[id_] = asyncio.Event()
        self._pending_reqs[(self, id_)] = (handler, time.time() + timeout)
        return  id_

    def send_notification(self, method, params=None):
        '''Send a notification.'''
        self.send_binary(self.notification_bytes(method, params))

    def send_notifications(self, mp_iterable):
        '''Send an iterable of (method, params) notification pairs.
        A 1-tuple is also valid in which case there are no params.'''
        if False and self.version.HAS_BATCHES:
            parts = [self.notification_bytes(*pair) for pair in mp_iterable]
            self.send_binary(self.version.batch_bytes(parts))
        else:
            for pair in mp_iterable:
                self.send_notification(*pair)

    # -- derived classes are intended to override these functions

    # Transport layer

    def is_closing(self):
        '''Return True if the underlying transport is closing.'''
        raise NotImplementedError

    def close_connection(self):
        '''Close the connection.'''
        raise NotImplementedError

    def send_bytes(self, binary):
        '''Pass the bytes through to the underlying transport.'''
        raise NotImplementedError

    # App layer

    def using_bandwidth(self, amount):
        '''Called as bandwidth is consumed.
        Override to implement bandwidth management.
        '''
        pass

    def notification_handler(self, method):
        '''Return the handler for the given notification.
        The handler can be synchronous or asynchronous.'''
        return None

    def request_handler(self, method):
        '''Return the handler for the given request method.
        The handler can be synchronous or asynchronous.'''
        return None


class JSONSession(JSONSessionBase, asyncio.Protocol):
    '''A JSONSessionBase instance specialized for use with
    asyncio.protocol to implement the transport layer.
    The app should await on items_event, which is set when unprocessed
    incoming items remain and cleared when the queue is empty, and
    then arrange to call process_pending_items asynchronously.
    Derived classes may want to override the request and notification
    handlers.
    '''

    def __init__(self, version=JSONRPCCompat):
        super().__init__(version=version)
        self.transport = None
        self.write_buffer_high = 500000

    def peer_info(self):
        '''Returns information about the peer.'''
        try:
            # get_extra_info can throw even if self.transport is not None
            return self.transport.get_extra_info('peername')
        except Exception:
            return None

    def abort(self):
        '''Cut the connection abruptly.'''
        self.transport.abort()

    def connection_made(self, transport):
        '''Handle an incoming client connection.'''
        transport.set_write_buffer_limits(high=self.write_buffer_high)
        self.transport = transport
        super().connection_made()

    def connection_lost(self, exc):
        '''Trigger timeouts of all pending requests.'''
        self.timeout_session()

    def is_closing(self):
        '''True if the underlying transport is closing.'''
        return self.transport and self.transport.is_closing()

    def close_connection(self):
        '''Close the connection.'''
        if self.transport:
            self.transport.close()

    def send_bytes(self, binary):
        '''Send JSON text over the transport.'''
        self.transport.writelines((binary, b'\n'))

    def peer_addr(self, anon=True):
        '''Return the peer address and port.'''
        peer_info = self.peer_info()
        if not peer_info:
            return 'unknown'
        if anon:
            return 'xx.xx.xx.xx:xx'
        if ':' in peer_info[0]:
            return '[{}]:{}'.format(peer_info[0], peer_info[1])
        else:
            return '{}:{}'.format(peer_info[0], peer_info[1])


#!/usr/bin/python
# -*- coding: utf-8 -*-
from enum import Enum


class TxType(Enum):
    """
    交易类型
    transaction type
    """
    NONE = 0
    REGISTER_ACCOUNT = 2
    COMMON = 3
    CONTRACT = 4
    CONTRACT_DEPLOY = 5
    DELEGATE = 6

    ASSET_ISSUE = 9
    ASSET_UPDATE = 10

    U_COIN_TRANSFER = 11
    U_CONTRACT_DEPLOY = 14
    U_CONTRACT_INVOKE = 15

    CDP_STAKE = 21
    CDP_REDEEM = 22
    CDP_LIQUIDATE = 23

    DEX_LIMITED_PRICE_BUY_ORDER = 84
    DEX_LIMITED_PRICE_SELL_ORDER = 85
    DEX_MARKET_PRICE_BUY_ORDER = 86
    DEX_MARKET_PRICE_SELL_ORDER = 87
    DEX_CANCEL_ORDER = 88


class VoteType(Enum):
    """
    投票类型
    """
    ADD = 1
    MINUS = 1


class AssetUpdateType(Enum):
    """

    """
    NONE = 0
    OWNER_UID = 1
    NAME = 2
    MINT_AMOUNT = 3


class CoinType(Enum):
    WICC ="WICC"
    WUSD = "WUSD"
    WGRT = "WGRT"
    WCNY = "WCNY"
    WBTC = "WBTC"
    WETH = "WETH"
    WEOS = "WEOS"
    USD = "USD"
    CNY = "CNY"
    EUR = "EUR"
    BTC = "BTC"
    USDT = "USDT"
    GOLD = "GOLD"
    KWH = "KWH"


class Transfer(object):

    def __init__(self, amount, symbol, desert_address):
        """
        :param amount: 支付的数量
        :param symbol: 支付的币种
        :param desert_address: 支付到的地址
        """
        self.pay_amount = amount
        self.pay_coin_symbol = symbol
        self.desert_address = desert_address


class Vote(object):

    def __init__(self, vote_type, pub_key, amount):
        """
        :param type: 投票类型
        :param pub_key: 接受投票账号的公钥
        :param amount: 投票的数量
        """
        self.type = vote_type
        self.pub_key = pub_key
        self.amount = amount


class BaseTransaction(object):
    """
    交易基类
    """
    def __init__(self):
        self.valid_height = 0
        self.register_id = ""
        self.fee_amount = 0
        self.fee_coin_symbol = ""
        self.memo = ""


class TransferTransaction(BaseTransaction):
    """
    转账交易
    """
    def __init__(self):
        super().__init__()
        self.transfer_list: [Transfer] = []


class VoteTransaction(BaseTransaction):
    """
    投票交易
    """
    def __init__(self):
        super().__init__()
        self.vote_list: [Vote] = []


class ContractDeployTransaction(BaseTransaction):
    """
    合约部署交易
    """
    def __init__(self):
        super().__init__()
        self.contract_byte = ""
        self.description = ""


class ContractCallTransaction(BaseTransaction):
    """
     合约调用交易
     """
    def __init__(self):
        super().__init__()
        self.app_id = ""
        self.contract_call_msg = ""
        self.pay_amount = 0
        self.pay_coin_symbol = ""


class ContractDeployTransaction(BaseTransaction):
    """
    合约部署
    """
    def __init__(self):
        super().__init__()
        self.file_path = ""


class DexLimitedPriceBuyTransaction(BaseTransaction):
    """
    限价买单交易
    """
    def __init__(self):
        super().__init__()
        self.fee_amount = 10000000
        self.coin_symbol = ""
        self.asset_symbol= ""
        self.asset_amount = ""
        self.price = 100000


class DexLimitedPriceSellTransaction(BaseTransaction):
    """
    限价卖单交易
    """
    def __init__(self):
        super().__init__()
        self.fee_amount = 10000000
        self.coin_symbol = ""
        self.asset_symbol= ""
        self.asset_amount = ""
        self.price = 100000


class DexMarketPriceBuyTransaction(BaseTransaction):
    """
    市价买单交易
    """
    def __init__(self):
        super().__init__()
        self.fee_amount = 10000000
        self.coin_symbol = ""
        self.coin_amount = 0
        self.asset_symbol= ""



class DexMarketPriceSellTransaction(BaseTransaction):
    """
    市价卖单交易
    """
    def __init__(self):
        super().__init__()
        self.fee_amount = 10000000
        self.coin_symbol = ""
        self.asset_symbol= ""
        self.asset_amount = ""


class DexCancelOrderTransaction(BaseTransaction):
    """
    取消挂单交易
    """
    def __init__(self):
        super().__init__()
        self.order_id = ""


class CdpStakeAsset(object):
    """
    Cdp抵押的资产
    """
    def __init__(self, symbol, amount):
        self.symbol = symbol
        self.amount = amount


class CdpStakeTransaction(BaseTransaction):
    """
    创建,追加cdp交易
    """
    def __init__(self):
        super().__init__()
        self.cdp_id = "0000000000000000000000000000000000000000000000000000000000000000"
        self.stake_list = []
        self.get_coin_symbol = ""
        self.get_amount = 0


class CdpRedeemAsset(object):
    """
    Cdp赎回的资产
    """
    def __init__(self, symbol, amount):
        self.symbol = symbol
        self.amount = amount


class CdpRedeemTransaction(BaseTransaction):
    """
    赎回cdp交易
    """
    def __init__(self):
        super().__init__()
        self.cdp_id = "0000000000000000000000000000000000000000000000000000000000000000"
        self.free_coin_symbol = ""
        self.repay_amount = 0
        self.free_amount = 0
        self.redeem_list = []


class CdpLiquidateTransaction(BaseTransaction):
    """
    清算cdp交易
    """
    def __init__(self):
        super().__init__()
        self.cdp_id = "0000000000000000000000000000000000000000000000000000000000000000"
        self.free_coin_symbol = ""
        self.liquidate_coin_symbol = 0
        self.liquidate_amount = 0


class AssetPublishTransaction(BaseTransaction):
    """
    资产发布交易
    """
    def __init__(self):
        super().__init__()
        self.asset_symbol = ""
        self.asset_name = ""
        self.asset_owner_regid = ""
        self.asset_amount = 0
        self.asset_renewable = True


class AssetUpdateTransaction(BaseTransaction):
    """
    资产更新交易
    """
    def __init__(self):
        super().__init__()
        self.asset_symbol = ""
        self.asset_update_type = 0
        self.asset_update_value = 0








## ripemd.py - pure Python implementation of the RIPEMD-160 algorithm.
## Bjorn Edstrom <be@bjrn.se> 16 december 2007.
##
## Copyrights
## ==========
##
## This code is a derived from an implementation by Markus Friedl which is
## subject to the following license. This Python implementation is not
## subject to any other license.
##
##/*
## * Copyright (c) 2001 Markus Friedl.  All rights reserved.
## *
## * Redistribution and use in source and binary forms, with or without
## * modification, are permitted provided that the following conditions
## * are met:
## * 1. Redistributions of source code must retain the above copyright
## *    notice, this list of conditions and the following disclaimer.
## * 2. Redistributions in binary form must reproduce the above copyright
## *    notice, this list of conditions and the following disclaimer in the
## *    documentation and/or other materials provided with the distribution.
## *
## * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
## * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
## * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
## * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
## * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
## * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
## * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
## */
##/*
## * Preneel, Bosselaers, Dobbertin, "The Cryptographic Hash Function RIPEMD-160",
## * RSA Laboratories, CryptoBytes, Volume 3, Number 2, Autumn 1997,
## * ftp://ftp.rsasecurity.com/pub/cryptobytes/crypto3n2.pdf
## */

try:
    import psyco
    psyco.full()
except ImportError:
    pass

import sys

is_python2 = sys.version_info.major == 2
#block_size = 1
digest_size = 20
digestsize = 20

try:
    range = xrange
except:
    pass

class RIPEMD160:
    """Return a new RIPEMD160 object. An optional string argument
    may be provided; if present, this string will be automatically
    hashed."""
    
    def __init__(self, arg=None):
        self.ctx = RMDContext()
        if arg:
            self.update(arg)
        self.dig = None
        
    def update(self, arg):
        """update(arg)"""        
        RMD160Update(self.ctx, arg, len(arg))
        self.dig = None
        
    def digest(self):
        """digest()"""        
        if self.dig:
            return self.dig
        ctx = self.ctx.copy()
        self.dig = RMD160Final(self.ctx)
        self.ctx = ctx
        return self.dig
    
    def hexdigest(self):
        """hexdigest()"""
        dig = self.digest()
        hex_digest = ''
        for d in dig:
            if (is_python2):
                hex_digest += '%02x' % ord(d)
            else:
                hex_digest += '%02x' % d
        return hex_digest
    
    def copy(self):
        """copy()"""        
        import copy
        return copy.deepcopy(self)



def new(arg=None):
    """Return a new RIPEMD160 object. An optional string argument
    may be provided; if present, this string will be automatically
    hashed."""    
    return RIPEMD160(arg)



#
# Private.
#

class RMDContext:
    def __init__(self):
        self.state = [0x67452301, 0xEFCDAB89, 0x98BADCFE,
                      0x10325476, 0xC3D2E1F0] # uint32
        self.count = 0 # uint64
        self.buffer = [0]*64 # uchar
    def copy(self):
        ctx = RMDContext()
        ctx.state = self.state[:]
        ctx.count = self.count
        ctx.buffer = self.buffer[:]
        return ctx

K0 = 0x00000000
K1 = 0x5A827999
K2 = 0x6ED9EBA1
K3 = 0x8F1BBCDC
K4 = 0xA953FD4E

KK0 = 0x50A28BE6
KK1 = 0x5C4DD124
KK2 = 0x6D703EF3
KK3 = 0x7A6D76E9
KK4 = 0x00000000

def ROL(n, x):
    return ((x << n) & 0xffffffff) | (x >> (32 - n))

def F0(x, y, z):
    return x ^ y ^ z

def F1(x, y, z):
    return (x & y) | (((~x) % 0x100000000) & z)

def F2(x, y, z):
    return (x | ((~y) % 0x100000000)) ^ z

def F3(x, y, z):
    return (x & z) | (((~z) % 0x100000000) & y)

def F4(x, y, z):
    return x ^ (y | ((~z) % 0x100000000))

def R(a, b, c, d, e, Fj, Kj, sj, rj, X):
    a = ROL(sj, (a + Fj(b, c, d) + X[rj] + Kj) % 0x100000000) + e
    c = ROL(10, c)
    return a % 0x100000000, c

PADDING = [0x80] + [0]*63

import sys
import struct

def RMD160Transform(state, block): #uint32 state[5], uchar block[64]
    x = [0]*16
    if sys.byteorder == 'little':
        if is_python2:
            x = struct.unpack('<16L', ''.join([chr(x) for x in block[0:64]]))
        else:
            x = struct.unpack('<16L', bytes(block[0:64]))
    else:
        raise "Error!!"
    a = state[0]
    b = state[1]
    c = state[2]
    d = state[3]
    e = state[4]

    #/* Round 1 */
    a, c = R(a, b, c, d, e, F0, K0, 11,  0, x);
    e, b = R(e, a, b, c, d, F0, K0, 14,  1, x);
    d, a = R(d, e, a, b, c, F0, K0, 15,  2, x);
    c, e = R(c, d, e, a, b, F0, K0, 12,  3, x);
    b, d = R(b, c, d, e, a, F0, K0,  5,  4, x);
    a, c = R(a, b, c, d, e, F0, K0,  8,  5, x);
    e, b = R(e, a, b, c, d, F0, K0,  7,  6, x);
    d, a = R(d, e, a, b, c, F0, K0,  9,  7, x);
    c, e = R(c, d, e, a, b, F0, K0, 11,  8, x);
    b, d = R(b, c, d, e, a, F0, K0, 13,  9, x);
    a, c = R(a, b, c, d, e, F0, K0, 14, 10, x);
    e, b = R(e, a, b, c, d, F0, K0, 15, 11, x);
    d, a = R(d, e, a, b, c, F0, K0,  6, 12, x);
    c, e = R(c, d, e, a, b, F0, K0,  7, 13, x);
    b, d = R(b, c, d, e, a, F0, K0,  9, 14, x);
    a, c = R(a, b, c, d, e, F0, K0,  8, 15, x); #/* #15 */
    #/* Round 2 */
    e, b = R(e, a, b, c, d, F1, K1,  7,  7, x);
    d, a = R(d, e, a, b, c, F1, K1,  6,  4, x);
    c, e = R(c, d, e, a, b, F1, K1,  8, 13, x);
    b, d = R(b, c, d, e, a, F1, K1, 13,  1, x);
    a, c = R(a, b, c, d, e, F1, K1, 11, 10, x);
    e, b = R(e, a, b, c, d, F1, K1,  9,  6, x);
    d, a = R(d, e, a, b, c, F1, K1,  7, 15, x);
    c, e = R(c, d, e, a, b, F1, K1, 15,  3, x);
    b, d = R(b, c, d, e, a, F1, K1,  7, 12, x);
    a, c = R(a, b, c, d, e, F1, K1, 12,  0, x);
    e, b = R(e, a, b, c, d, F1, K1, 15,  9, x);
    d, a = R(d, e, a, b, c, F1, K1,  9,  5, x);
    c, e = R(c, d, e, a, b, F1, K1, 11,  2, x);
    b, d = R(b, c, d, e, a, F1, K1,  7, 14, x);
    a, c = R(a, b, c, d, e, F1, K1, 13, 11, x);
    e, b = R(e, a, b, c, d, F1, K1, 12,  8, x); #/* #31 */
    #/* Round 3 */
    d, a = R(d, e, a, b, c, F2, K2, 11,  3, x);
    c, e = R(c, d, e, a, b, F2, K2, 13, 10, x);
    b, d = R(b, c, d, e, a, F2, K2,  6, 14, x);
    a, c = R(a, b, c, d, e, F2, K2,  7,  4, x);
    e, b = R(e, a, b, c, d, F2, K2, 14,  9, x);
    d, a = R(d, e, a, b, c, F2, K2,  9, 15, x);
    c, e = R(c, d, e, a, b, F2, K2, 13,  8, x);
    b, d = R(b, c, d, e, a, F2, K2, 15,  1, x);
    a, c = R(a, b, c, d, e, F2, K2, 14,  2, x);
    e, b = R(e, a, b, c, d, F2, K2,  8,  7, x);
    d, a = R(d, e, a, b, c, F2, K2, 13,  0, x);
    c, e = R(c, d, e, a, b, F2, K2,  6,  6, x);
    b, d = R(b, c, d, e, a, F2, K2,  5, 13, x);
    a, c = R(a, b, c, d, e, F2, K2, 12, 11, x);
    e, b = R(e, a, b, c, d, F2, K2,  7,  5, x);
    d, a = R(d, e, a, b, c, F2, K2,  5, 12, x); #/* #47 */
    #/* Round 4 */
    c, e = R(c, d, e, a, b, F3, K3, 11,  1, x);
    b, d = R(b, c, d, e, a, F3, K3, 12,  9, x);
    a, c = R(a, b, c, d, e, F3, K3, 14, 11, x);
    e, b = R(e, a, b, c, d, F3, K3, 15, 10, x);
    d, a = R(d, e, a, b, c, F3, K3, 14,  0, x);
    c, e = R(c, d, e, a, b, F3, K3, 15,  8, x);
    b, d = R(b, c, d, e, a, F3, K3,  9, 12, x);
    a, c = R(a, b, c, d, e, F3, K3,  8,  4, x);
    e, b = R(e, a, b, c, d, F3, K3,  9, 13, x);
    d, a = R(d, e, a, b, c, F3, K3, 14,  3, x);
    c, e = R(c, d, e, a, b, F3, K3,  5,  7, x);
    b, d = R(b, c, d, e, a, F3, K3,  6, 15, x);
    a, c = R(a, b, c, d, e, F3, K3,  8, 14, x);
    e, b = R(e, a, b, c, d, F3, K3,  6,  5, x);
    d, a = R(d, e, a, b, c, F3, K3,  5,  6, x);
    c, e = R(c, d, e, a, b, F3, K3, 12,  2, x); #/* #63 */
    #/* Round 5 */
    b, d = R(b, c, d, e, a, F4, K4,  9,  4, x);
    a, c = R(a, b, c, d, e, F4, K4, 15,  0, x);
    e, b = R(e, a, b, c, d, F4, K4,  5,  5, x);
    d, a = R(d, e, a, b, c, F4, K4, 11,  9, x);
    c, e = R(c, d, e, a, b, F4, K4,  6,  7, x);
    b, d = R(b, c, d, e, a, F4, K4,  8, 12, x);
    a, c = R(a, b, c, d, e, F4, K4, 13,  2, x);
    e, b = R(e, a, b, c, d, F4, K4, 12, 10, x);
    d, a = R(d, e, a, b, c, F4, K4,  5, 14, x);
    c, e = R(c, d, e, a, b, F4, K4, 12,  1, x);
    b, d = R(b, c, d, e, a, F4, K4, 13,  3, x);
    a, c = R(a, b, c, d, e, F4, K4, 14,  8, x);
    e, b = R(e, a, b, c, d, F4, K4, 11, 11, x);
    d, a = R(d, e, a, b, c, F4, K4,  8,  6, x);
    c, e = R(c, d, e, a, b, F4, K4,  5, 15, x);
    b, d = R(b, c, d, e, a, F4, K4,  6, 13, x); #/* #79 */

    aa = a;
    bb = b;
    cc = c;
    dd = d;
    ee = e;

    a = state[0]
    b = state[1]
    c = state[2]
    d = state[3]
    e = state[4]    

    #/* Parallel round 1 */
    a, c = R(a, b, c, d, e, F4, KK0,  8,  5, x)
    e, b = R(e, a, b, c, d, F4, KK0,  9, 14, x)
    d, a = R(d, e, a, b, c, F4, KK0,  9,  7, x)
    c, e = R(c, d, e, a, b, F4, KK0, 11,  0, x)
    b, d = R(b, c, d, e, a, F4, KK0, 13,  9, x)
    a, c = R(a, b, c, d, e, F4, KK0, 15,  2, x)
    e, b = R(e, a, b, c, d, F4, KK0, 15, 11, x)
    d, a = R(d, e, a, b, c, F4, KK0,  5,  4, x)
    c, e = R(c, d, e, a, b, F4, KK0,  7, 13, x)
    b, d = R(b, c, d, e, a, F4, KK0,  7,  6, x)
    a, c = R(a, b, c, d, e, F4, KK0,  8, 15, x)
    e, b = R(e, a, b, c, d, F4, KK0, 11,  8, x)
    d, a = R(d, e, a, b, c, F4, KK0, 14,  1, x)
    c, e = R(c, d, e, a, b, F4, KK0, 14, 10, x)
    b, d = R(b, c, d, e, a, F4, KK0, 12,  3, x)
    a, c = R(a, b, c, d, e, F4, KK0,  6, 12, x) #/* #15 */
    #/* Parallel round 2 */
    e, b = R(e, a, b, c, d, F3, KK1,  9,  6, x)
    d, a = R(d, e, a, b, c, F3, KK1, 13, 11, x)
    c, e = R(c, d, e, a, b, F3, KK1, 15,  3, x)
    b, d = R(b, c, d, e, a, F3, KK1,  7,  7, x)
    a, c = R(a, b, c, d, e, F3, KK1, 12,  0, x)
    e, b = R(e, a, b, c, d, F3, KK1,  8, 13, x)
    d, a = R(d, e, a, b, c, F3, KK1,  9,  5, x)
    c, e = R(c, d, e, a, b, F3, KK1, 11, 10, x)
    b, d = R(b, c, d, e, a, F3, KK1,  7, 14, x)
    a, c = R(a, b, c, d, e, F3, KK1,  7, 15, x)
    e, b = R(e, a, b, c, d, F3, KK1, 12,  8, x)
    d, a = R(d, e, a, b, c, F3, KK1,  7, 12, x)
    c, e = R(c, d, e, a, b, F3, KK1,  6,  4, x)
    b, d = R(b, c, d, e, a, F3, KK1, 15,  9, x)
    a, c = R(a, b, c, d, e, F3, KK1, 13,  1, x)
    e, b = R(e, a, b, c, d, F3, KK1, 11,  2, x) #/* #31 */
    #/* Parallel round 3 */
    d, a = R(d, e, a, b, c, F2, KK2,  9, 15, x)
    c, e = R(c, d, e, a, b, F2, KK2,  7,  5, x)
    b, d = R(b, c, d, e, a, F2, KK2, 15,  1, x)
    a, c = R(a, b, c, d, e, F2, KK2, 11,  3, x)
    e, b = R(e, a, b, c, d, F2, KK2,  8,  7, x)
    d, a = R(d, e, a, b, c, F2, KK2,  6, 14, x)
    c, e = R(c, d, e, a, b, F2, KK2,  6,  6, x)
    b, d = R(b, c, d, e, a, F2, KK2, 14,  9, x)
    a, c = R(a, b, c, d, e, F2, KK2, 12, 11, x)
    e, b = R(e, a, b, c, d, F2, KK2, 13,  8, x)
    d, a = R(d, e, a, b, c, F2, KK2,  5, 12, x)
    c, e = R(c, d, e, a, b, F2, KK2, 14,  2, x)
    b, d = R(b, c, d, e, a, F2, KK2, 13, 10, x)
    a, c = R(a, b, c, d, e, F2, KK2, 13,  0, x)
    e, b = R(e, a, b, c, d, F2, KK2,  7,  4, x)
    d, a = R(d, e, a, b, c, F2, KK2,  5, 13, x) #/* #47 */
    #/* Parallel round 4 */
    c, e = R(c, d, e, a, b, F1, KK3, 15,  8, x)
    b, d = R(b, c, d, e, a, F1, KK3,  5,  6, x)
    a, c = R(a, b, c, d, e, F1, KK3,  8,  4, x)
    e, b = R(e, a, b, c, d, F1, KK3, 11,  1, x)
    d, a = R(d, e, a, b, c, F1, KK3, 14,  3, x)
    c, e = R(c, d, e, a, b, F1, KK3, 14, 11, x)
    b, d = R(b, c, d, e, a, F1, KK3,  6, 15, x)
    a, c = R(a, b, c, d, e, F1, KK3, 14,  0, x)
    e, b = R(e, a, b, c, d, F1, KK3,  6,  5, x)
    d, a = R(d, e, a, b, c, F1, KK3,  9, 12, x)
    c, e = R(c, d, e, a, b, F1, KK3, 12,  2, x)
    b, d = R(b, c, d, e, a, F1, KK3,  9, 13, x)
    a, c = R(a, b, c, d, e, F1, KK3, 12,  9, x)
    e, b = R(e, a, b, c, d, F1, KK3,  5,  7, x)
    d, a = R(d, e, a, b, c, F1, KK3, 15, 10, x)
    c, e = R(c, d, e, a, b, F1, KK3,  8, 14, x) #/* #63 */
    #/* Parallel round 5 */
    b, d = R(b, c, d, e, a, F0, KK4,  8, 12, x)
    a, c = R(a, b, c, d, e, F0, KK4,  5, 15, x)
    e, b = R(e, a, b, c, d, F0, KK4, 12, 10, x)
    d, a = R(d, e, a, b, c, F0, KK4,  9,  4, x)
    c, e = R(c, d, e, a, b, F0, KK4, 12,  1, x)
    b, d = R(b, c, d, e, a, F0, KK4,  5,  5, x)
    a, c = R(a, b, c, d, e, F0, KK4, 14,  8, x)
    e, b = R(e, a, b, c, d, F0, KK4,  6,  7, x)
    d, a = R(d, e, a, b, c, F0, KK4,  8,  6, x)
    c, e = R(c, d, e, a, b, F0, KK4, 13,  2, x)
    b, d = R(b, c, d, e, a, F0, KK4,  6, 13, x)
    a, c = R(a, b, c, d, e, F0, KK4,  5, 14, x)
    e, b = R(e, a, b, c, d, F0, KK4, 15,  0, x)
    d, a = R(d, e, a, b, c, F0, KK4, 13,  3, x)
    c, e = R(c, d, e, a, b, F0, KK4, 11,  9, x)
    b, d = R(b, c, d, e, a, F0, KK4, 11, 11, x) #/* #79 */

    t = (state[1] + cc + d) % 0x100000000;
    state[1] = (state[2] + dd + e) % 0x100000000;
    state[2] = (state[3] + ee + a) % 0x100000000;
    state[3] = (state[4] + aa + b) % 0x100000000;
    state[4] = (state[0] + bb + c) % 0x100000000;
    state[0] = t % 0x100000000;

    pass


def RMD160Update(ctx, inp, inplen):
    if type(inp) == str:
        inp = [ord(i)&0xff for i in inp]
    
    have = int((ctx.count // 8) % 64)
    inplen = int(inplen)
    need = 64 - have
    ctx.count += 8 * inplen
    off = 0
    if inplen >= need:
        if have:
            for i in range(need):
                ctx.buffer[have+i] = inp[i]
            RMD160Transform(ctx.state, ctx.buffer)
            off = need
            have = 0
        while off + 64 <= inplen:
            RMD160Transform(ctx.state, inp[off:]) #<---
            off += 64
    if off < inplen:
        # memcpy(ctx->buffer + have, input+off, len-off);
        for i in range(inplen - off):
            ctx.buffer[have+i] = inp[off+i]

def RMD160Final(ctx):
    size = struct.pack("<Q", ctx.count)
    padlen = 64 - ((ctx.count // 8) % 64)
    if padlen < 1+8:
        padlen += 64
    RMD160Update(ctx, PADDING, padlen-8)
    RMD160Update(ctx, size, 8)
    return struct.pack("<5L", *ctx.state)


assert '37f332f68db77bd9d7edd4969571ad671cf9dd3b' == \
       new('The quick brown fox jumps over the lazy dog').hexdigest()
assert '132072df690933835eb8b6ad0b77e7b6f14acad7' == \
       new('The quick brown fox jumps over the lazy cog').hexdigest()
assert '9c1185a5c5e9fc54612808977ee8f548b2258d31' == \
       new('').hexdigest()



import re
import requests
import datetime
from .utils import parse_addr_args

#Base module for all insight-based explorers

sendtx_url = "%s/tx/send"
address_url = "%s/addrs/%s/txs"
utxo_url = "%s/addrs/%s/utxo"
fetchtx_url = "%s/tx/%s"
current_block_height_url = "%s/status?q=getInfo"
block_hash_by_height_url = "%s/block-index/%s"
block_info_url = "%s/block/%s"

def unspent(base_url, *args):

    addrs = parse_addr_args(*args)

    if len(addrs) == 0:
        return []

    url = utxo_url % (base_url, ','.join(addrs))

    response = requests.get(url)
    txs = response.json()
    for i, tx in enumerate(txs):
        if 'satoshis' in tx.keys():
            txs[i] = {
                'output': "%s:%s" % (tx['txid'], tx['vout']),
                'value': tx['satoshis'],
            }
        else:
            txs[i] = {
                'output': "%s:%s" % (tx['txid'], tx['vout']),
                'value': int(tx['amount'] * 100000000),
            }
    return txs

def fetchtx(base_url, txhash):
    url = fetchtx_url % (base_url, txhash)
    response = requests.get(url)
    return response.json()

def txinputs(base_url, txhash):
    result = fetchtx(base_url, txhash)
    inputs = result['vin']
    unspents = [{'output': "%s:%s" % (i['txid'], i['vout']), 'value': i['valueSat']} for i in inputs]
    return unspents

def pushtx(base_url, network, tx):
    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')

    url = sendtx_url % base_url
    response = requests.post(url, {'rawtx': tx})
    if response.status_code == 200:
        result = response.json()
        return {'status': 'success',
                'data': {
                    'txid': result['txid'],
                    'network': network
                    }
                }
    return response

# Gets the transaction output history of a given set of addresses,
# including whether or not they have been spent
def history(base_url, *args):
    # Valid input formats: history([addr1, addr2,addr3])
    #                      history(addr1, addr2, addr3)

    addrs = parse_addr_args(*args)

    if len(addrs) == 0:
        return []

    url = address_url % (base_url, ','.join(addrs))
    response = requests.get(url)
    return response.json()

def block_height(base_url, txhash):
    tx = fetchtx(base_url, txhash)
    return tx.get('blockheight', None) or tx.get('height', None)

def block_info(base_url, height):
    url = block_hash_by_height_url % (base_url, height)
    response = requests.get(url)
    blockhash = response.json()['blockHash']
    url = block_info_url % (base_url, blockhash)
    response = requests.get(url)
    data = response.json()
    return {
        'version': data['version'],
        'hash': data['hash'],
        'prevhash': data['previousblockhash'],
        'timestamp': data['time'],
        'merkle_root': data['merkleroot'],
        'bits': data['bits'],
        'nonce': data['nonce'],
        'tx_hashes': data['tx']
    }


def current_block_height(base_url):
    url = current_block_height_url % base_url
    response = requests.get(url)
    result = response.json()
    return result['info']['blocks']


# coding=utf-8
from wicc.wiccrpc.box.RpcManager import RpcManager


class Contract(object):

    @classmethod
    def register_contract(cls, address, file_path, fee=110000000):
        """
        用于发布智能合约
        :param address: 发布智能合约者的地址
        :param file_path: 发布的智能合约文件路径+文件名
        :param fee: 发布智能合约所需费用 （一般至少110000000
        :return: 发布合约交易的哈希
        """
        data = {
            "method": "registercontracttx",
            "params": [address, file_path, fee]
        }
        return RpcManager.request(data)

    @classmethod
    def contract_regid(cls, contract_hash):
        """
        获取智能合约的regid
        :param contract_hash: 发布合约时的交易哈希
        :return: 该智能合约的regid
        """
        data = {
            "method": "getcontractregid",
            "params": [contract_hash]
        }
        return RpcManager.request(data)['result']

    @classmethod
    def contract_info(cls, contract_id):
        """
        获取智能合约信息
        :param contract_id: 智能合约 regid
        :return: 智能合约信息
        """
        data = {
            "method": "getcontractinfo",
            "params": [contract_id]
        }
        return RpcManager.request(data)

    @classmethod
    def contract_account_info(cls, contract_regid, user):
        """
        用户在智能合约中的相关信息
        :param contract_regid: 发布合约时的交易哈希
        :param user: 查询的用户地址/regid
        :return: 该智能合约的regid
        """
        data = {
            "method": "getcontractaccountinfo",
            "params": [contract_regid, user]
        }
        return RpcManager.request(data)['result']

    @classmethod
    def contract_data(cls, contract_regid, key):
        """
        获取智能合约相关原生数据信息
        :param contract_regid: 合约regid
        :param key: 智能合约数据的key值
        :return: 该智能合约数据
        """
        data = {
            "method": "getcontractdata",
            "params": [contract_regid, key]
        }
        return RpcManager.request(data)['result']

    @classmethod
    def call_contract(cls, user, contract_regid, contract_command, fee=1000000, amount=0,):
        """
        调用智能合约(合约必须是被确认状态)
        :param user: 合约调用者地址/regid
        :param contract_regid: 智能合约的regid
        :param amount: 向智能合约发送维基币的数量(sawi)
        :param contract_command: 调用合约的命令
        :param fee: 调用合约交易所需费用
        :return: 调用合约的交易哈希
        """
        data = {
            "method": "callcontracttx",
            "params": [user, contract_regid, amount, contract_command, fee]
        }
        return RpcManager.request(data)

    @classmethod
    def contract_list(cls, show_detail=False):
        """
        获取已发布的智能合约列表
        :param show_detail:
        :return:
        """
        data = {
            "method": "listcontracts",
            "params": [show_detail]
        }
        return RpcManager.request(data)['result']




ELECTRUM_VERSION = '3.0.5'   # version of the client package
PROTOCOL_VERSION = '1.1'     # protocol version requested

# The hash of the mnemonic seed must begin with this
SEED_PREFIX      = '01'      # Standard wallet
SEED_PREFIX_2FA  = '101'     # Two-factor authentication
SEED_PREFIX_SW   = '100'     # Segwit wallet



#!/usr/bin/python
from .py2specials import *
from .py3specials import *
import binascii
import hashlib
import re
import base64
import time
import random
import hmac
from .ripemd import *

# Elliptic curve parameters (secp256k1)

P = 2**256 - 2**32 - 977
N = 115792089237316195423570985008687907852837564279074904382605163141518161494337
A = 0
B = 7
Gx = 55066263022277343669578718895168534326250603453777594175500187360389116729240
Gy = 32670510020758816978083085130507043184471273380659243275938904335757337482424
G = (Gx, Gy)


def change_curve(p, n, a, b, gx, gy):
    global P, N, A, B, Gx, Gy, G
    P, N, A, B, Gx, Gy = p, n, a, b, gx, gy
    G = (Gx, Gy)


def getG():
    return G

# Extended Euclidean Algorithm


def inv(a, n):
    if a == 0:
        return 0
    lm, hm = 1, 0
    low, high = a % n, n
    while low > 1:
        r = high//low
        nm, new = hm-lm*r, high-low*r
        lm, low, hm, high = nm, new, lm, low
    return lm % n



# JSON access (for pybtctool convenience)


def access(obj, prop):
    if isinstance(obj, dict):
        if prop in obj:
            return obj[prop]
        elif '.' in prop:
            return obj[float(prop)]
        else:
            return obj[int(prop)]
    else:
        return obj[int(prop)]


def multiaccess(obj, prop):
    return [access(o, prop) for o in obj]


def slice(obj, start=0, end=2**200):
    return obj[int(start):int(end)]


def count(obj):
    return len(obj)

_sum = sum


def sum(obj):
    return _sum(obj)


def isinf(p):
    return p[0] == 0 and p[1] == 0


def to_jacobian(p):
    o = (p[0], p[1], 1)
    return o


def jacobian_double(p):
    if not p[1]:
        return (0, 0, 0)
    ysq = (p[1] ** 2) % P
    S = (4 * p[0] * ysq) % P
    M = (3 * p[0] ** 2 + A * p[2] ** 4) % P
    nx = (M**2 - 2 * S) % P
    ny = (M * (S - nx) - 8 * ysq ** 2) % P
    nz = (2 * p[1] * p[2]) % P
    return (nx, ny, nz)


def jacobian_add(p, q):
    if not p[1]:
        return q
    if not q[1]:
        return p
    U1 = (p[0] * q[2] ** 2) % P
    U2 = (q[0] * p[2] ** 2) % P
    S1 = (p[1] * q[2] ** 3) % P
    S2 = (q[1] * p[2] ** 3) % P
    if U1 == U2:
        if S1 != S2:
            return (0, 0, 1)
        return jacobian_double(p)
    H = U2 - U1
    R = S2 - S1
    H2 = (H * H) % P
    H3 = (H * H2) % P
    U1H2 = (U1 * H2) % P
    nx = (R ** 2 - H3 - 2 * U1H2) % P
    ny = (R * (U1H2 - nx) - S1 * H3) % P
    nz = (H * p[2] * q[2]) % P
    return (nx, ny, nz)


def from_jacobian(p):
    z = inv(p[2], P)
    return ((p[0] * z**2) % P, (p[1] * z**3) % P)


def jacobian_multiply(a, n):
    if a[1] == 0 or n == 0:
        return (0, 0, 1)
    if n == 1:
        return a
    if n < 0 or n >= N:
        return jacobian_multiply(a, n % N)
    if (n % 2) == 0:
        return jacobian_double(jacobian_multiply(a, n//2))
    if (n % 2) == 1:
        return jacobian_add(jacobian_double(jacobian_multiply(a, n//2)), a)


def fast_multiply(a, n):
    return from_jacobian(jacobian_multiply(to_jacobian(a), n))


def fast_add(a, b):
    return from_jacobian(jacobian_add(to_jacobian(a), to_jacobian(b)))

# Functions for handling pubkey and privkey formats


def get_pubkey_format(pub):
    if is_python2:
        two = '\x02'
        three = '\x03'
        four = '\x04'
    else:
        two = 2
        three = 3
        four = 4

    if isinstance(pub, (tuple, list)): return 'decimal'
    elif len(pub) == 65 and pub[0] == four: return 'bin'
    elif len(pub) == 130 and pub[0:2] == '04': return 'hex'
    elif len(pub) == 33 and pub[0] in [two, three]: return 'bin_compressed'
    elif len(pub) == 66 and pub[0:2] in ['02', '03']: return 'hex_compressed'
    elif len(pub) == 64: return 'bin_electrum'
    elif len(pub) == 128: return 'hex_electrum'
    else: raise Exception("Pubkey not in recognized format")


def encode_pubkey(pub, formt):
    if not isinstance(pub, (tuple, list)):
        pub = decode_pubkey(pub)
    if formt == 'decimal': return pub
    elif formt == 'bin': return b'\x04' + encode(pub[0], 256, 32) + encode(pub[1], 256, 32)
    elif formt == 'bin_compressed':
        return from_int_to_byte(2+(pub[1] % 2)) + encode(pub[0], 256, 32)
    elif formt == 'hex': return '04' + encode(pub[0], 16, 64) + encode(pub[1], 16, 64)
    elif formt == 'hex_compressed':
        return '0'+str(2+(pub[1] % 2)) + encode(pub[0], 16, 64)
    elif formt == 'bin_electrum': return encode(pub[0], 256, 32) + encode(pub[1], 256, 32)
    elif formt == 'hex_electrum': return encode(pub[0], 16, 64) + encode(pub[1], 16, 64)
    else: raise Exception("Invalid format!")


def decode_pubkey(pub, formt=None):
    if not formt: formt = get_pubkey_format(pub)
    if formt == 'decimal': return pub
    elif formt == 'bin': return (decode(pub[1:33], 256), decode(pub[33:65], 256))
    elif formt == 'bin_compressed':
        x = decode(pub[1:33], 256)
        beta = pow(int(x*x*x+A*x+B), int((P+1)//4), int(P))
        y = (P-beta) if ((beta + from_byte_to_int(pub[0])) % 2) else beta
        return (x, y)
    elif formt == 'hex': return (decode(pub[2:66], 16), decode(pub[66:130], 16))
    elif formt == 'hex_compressed':
        return decode_pubkey(safe_from_hex(pub), 'bin_compressed')
    elif formt == 'bin_electrum':
        return (decode(pub[:32], 256), decode(pub[32:64], 256))
    elif formt == 'hex_electrum':
        return (decode(pub[:64], 16), decode(pub[64:128], 16))
    else: raise Exception("Invalid format!")

def get_privkey_format(priv):
    if isinstance(priv, int_types): return 'decimal'
    elif len(priv) == 32: return 'bin'
    elif len(priv) == 33: return 'bin_compressed'
    elif len(priv) == 64: return 'hex'
    elif len(priv) == 66: return 'hex_compressed'
    else:
        bin_p = b58check_to_bin(priv)
        if len(bin_p) == 32: return 'wif'
        elif len(bin_p) == 33: return 'wif_compressed'
        else: raise Exception("WIF does not represent privkey")

def encode_privkey(priv, formt, vbyte=128):
    if not isinstance(priv, int_types):
        return encode_privkey(decode_privkey(priv), formt, vbyte)
    if formt == 'decimal': return priv
    elif formt == 'bin': return encode(priv, 256, 32)
    elif formt == 'bin_compressed': return encode(priv, 256, 32)+b'\x01'
    elif formt == 'hex': return encode(priv, 16, 64)
    elif formt == 'hex_compressed': return encode(priv, 16, 64)+'01'
    elif formt == 'wif':
        return bin_to_b58check(encode(priv, 256, 32), int(vbyte))
    elif formt == 'wif_compressed':
        return bin_to_b58check(encode(priv, 256, 32) + b'\x01', int(vbyte))
    else: raise Exception("Invalid format!")

def decode_privkey(priv,formt=None):
    if not formt: formt = get_privkey_format(priv)
    if formt == 'decimal': return priv
    elif formt == 'bin': return decode(priv, 256)
    elif formt == 'bin_compressed': return decode(priv[:32], 256)
    elif formt == 'hex': return decode(priv, 16)
    elif formt == 'hex_compressed': return decode(priv[:64], 16)
    elif formt == 'wif': return decode(b58check_to_bin(priv),256)
    elif formt == 'wif_compressed':
        return decode(b58check_to_bin(priv)[:32],256)
    else: raise Exception("WIF does not represent privkey")

def add_pubkeys(p1, p2):
    f1, f2 = get_pubkey_format(p1), get_pubkey_format(p2)
    return encode_pubkey(fast_add(decode_pubkey(p1, f1), decode_pubkey(p2, f2)), f1)

def add_privkeys(p1, p2):
    f1, f2 = get_privkey_format(p1), get_privkey_format(p2)
    return encode_privkey((decode_privkey(p1, f1) + decode_privkey(p2, f2)) % N, f1)


def mul_privkeys(p1, p2):
    f1, f2 = get_privkey_format(p1), get_privkey_format(p2)
    return encode_privkey((decode_privkey(p1, f1) * decode_privkey(p2, f2)) % N, f1)

def multiply(pubkey, privkey):
    f1, f2 = get_pubkey_format(pubkey), get_privkey_format(privkey)
    pubkey, privkey = decode_pubkey(pubkey, f1), decode_privkey(privkey, f2)
    # http://safecurves.cr.yp.to/twist.html
    if not isinf(pubkey) and (pubkey[0]**3+B-pubkey[1]*pubkey[1]) % P != 0:
        raise Exception("Point not on curve")
    return encode_pubkey(fast_multiply(pubkey, privkey), f1)


def divide(pubkey, privkey):
    factor = inv(decode_privkey(privkey), N)
    return multiply(pubkey, factor)


def compress(pubkey):
    f = get_pubkey_format(pubkey)
    if 'compressed' in f: return pubkey
    elif f == 'bin': return encode_pubkey(decode_pubkey(pubkey, f), 'bin_compressed')
    elif f == 'hex' or f == 'decimal':
        return encode_pubkey(decode_pubkey(pubkey, f), 'hex_compressed')


def decompress(pubkey):
    f = get_pubkey_format(pubkey)
    if 'compressed' not in f: return pubkey
    elif f == 'bin_compressed': return encode_pubkey(decode_pubkey(pubkey, f), 'bin')
    elif f == 'hex_compressed' or f == 'decimal':
        return encode_pubkey(decode_pubkey(pubkey, f), 'hex')


def privkey_to_pubkey(privkey):
    f = get_privkey_format(privkey)
    privkey = decode_privkey(privkey, f)
    if privkey >= N:
        raise Exception("Invalid privkey")
    if f in ['bin', 'bin_compressed', 'hex', 'hex_compressed', 'decimal']:
        return encode_pubkey(fast_multiply(G, privkey), f)
    else:
        return encode_pubkey(fast_multiply(G, privkey), f.replace('wif', 'hex'))

privtopub = privkey_to_pubkey


def privkey_to_address(priv, magicbyte=0):
    return pubkey_to_address(privkey_to_pubkey(priv), magicbyte)
privtoaddr = privkey_to_address


def neg_pubkey(pubkey):
    f = get_pubkey_format(pubkey)
    pubkey = decode_pubkey(pubkey, f)
    return encode_pubkey((pubkey[0], (P-pubkey[1]) % P), f)


def neg_privkey(privkey):
    f = get_privkey_format(privkey)
    privkey = decode_privkey(privkey, f)
    return encode_privkey((N - privkey) % N, f)

def subtract_pubkeys(p1, p2):
    f1, f2 = get_pubkey_format(p1), get_pubkey_format(p2)
    k2 = decode_pubkey(p2, f2)
    return encode_pubkey(fast_add(decode_pubkey(p1, f1), (k2[0], (P - k2[1]) % P)), f1)


def subtract_privkeys(p1, p2):
    f1, f2 = get_privkey_format(p1), get_privkey_format(p2)
    k2 = decode_privkey(p2, f2)
    return encode_privkey((decode_privkey(p1, f1) - k2) % N, f1)



# Hashes


def bin_hash160(string):
    intermed = hashlib.sha256(string).digest()
    digest = ''
    try:
        digest = hashlib.new('ripemd160', intermed).digest()
    except:
        digest = RIPEMD160(intermed).digest()
    return digest

def hash160(string):
    return safe_hexlify(bin_hash160(string))

def hex_to_hash160(s_hex):
    return hash160(binascii.unhexlify(s_hex))

def bin_sha256(string):
    binary_data = string if isinstance(string, bytes) else bytes(string, 'utf-8')
    return hashlib.sha256(binary_data).digest()

def sha256(string):
    return bytes_to_hex_string(bin_sha256(string))


def bin_ripemd160(string):
    try:
        digest = hashlib.new('ripemd160', string).digest()
    except:
        digest = RIPEMD160(string).digest()
    return digest


def ripemd160(string):
    return safe_hexlify(bin_ripemd160(string))


def bin_dbl_sha256(s):
    bytes_to_hash = from_string_to_bytes(s)
    return hashlib.sha256(hashlib.sha256(bytes_to_hash).digest()).digest()


def dbl_sha256(string):
    return safe_hexlify(bin_dbl_sha256(string))


def bin_slowsha(string):
    string = from_string_to_bytes(string)
    orig_input = string
    for i in range(100000):
        string = hashlib.sha256(string + orig_input).digest()
    return string


def slowsha(string):
    return safe_hexlify(bin_slowsha(string))


def hash_to_int(x):
    if len(x) in [40, 64]:
        return decode(x, 16)
    return decode(x, 256)


def num_to_var_int(x):
    x = int(x)
    if x < 253: return from_int_to_byte(x)
    elif x < 65536: return from_int_to_byte(253)+encode(x, 256, 2)[::-1]
    elif x < 4294967296: return from_int_to_byte(254) + encode(x, 256, 4)[::-1]
    else: return from_int_to_byte(255) + encode(x, 256, 8)[::-1]


# WTF, Electrum?
def electrum_sig_hash(message):
    padded = b"\x18Bitcoin Signed Message:\n" + num_to_var_int(len(message)) + from_string_to_bytes(message)
    return bin_dbl_sha256(padded)


def random_key():
    # Gotta be secure after that java.SecureRandom fiasco...
    entropy = random_string(32) \
        + str(random.randrange(2**256)) \
        + str(int(time.time() * 1000000))
    return sha256(entropy)


def random_electrum_seed():
    #entropy = os.urandom(32) \ # fails in Python 3, hence copied from random_key()
    entropy = random_string(32) \
        + str(random.randrange(2**256)) \
        + str(int(time.time() * 1000000))
    return sha256(entropy)[:32]

# Encodings

def b58check_to_bin(inp):
    leadingzbytes = len(re.match('^1*', inp).group(0))
    data = b'\x00' * leadingzbytes + changebase(inp, 58, 256)
    assert bin_dbl_sha256(data[:-4])[:4] == data[-4:]
    return data[1:-4]


def get_version_byte(inp):
    leadingzbytes = len(re.match('^1*', inp).group(0))
    data = b'\x00' * leadingzbytes + changebase(inp, 58, 256)
    assert bin_dbl_sha256(data[:-4])[:4] == data[-4:]
    return ord(data[0])


def hex_to_b58check(inp, magicbyte=0):
    return bin_to_b58check(binascii.unhexlify(inp), magicbyte)


def b58check_to_hex(inp):
    return safe_hexlify(b58check_to_bin(inp))

def pubkey_to_hash(pubkey):
    if isinstance(pubkey, (list, tuple)):
        pubkey = encode_pubkey(pubkey, 'bin')
    if len(pubkey) in [66, 130]:
        return bin_hash160(binascii.unhexlify(pubkey))
    return bin_hash160(pubkey)

def pubkey_to_hash_hex(pubkey):
    return safe_hexlify(pubkey_to_hash(pubkey))

def pubkey_to_address(pubkey, magicbyte=0):
    pubkey_hash = pubkey_to_hash(pubkey)
    return bin_to_b58check(pubkey_hash, magicbyte)

pubtoaddr = pubkey_to_address


def is_privkey(priv):
    try:
        get_privkey_format(priv)
        return True
    except:
        return False

def is_pubkey(pubkey):
    try:
        get_pubkey_format(pubkey)
        return True
    except:
        return False


# EDCSA


def encode_sig(v, r, s):
    vb, rb, sb = from_int_to_byte(v), encode(r, 256), encode(s, 256)
    
    result = base64.b64encode(vb+b'\x00'*(32-len(rb))+rb+b'\x00'*(32-len(sb))+sb)
    return result if is_python2 else str(result, 'utf-8')


def decode_sig(sig):
    bytez = base64.b64decode(sig)
    return from_byte_to_int(bytez[0]), decode(bytez[1:33], 256), decode(bytez[33:], 256)

# https://tools.ietf.org/html/rfc6979#section-3.2


def deterministic_generate_k(msghash, priv):
    v = b'\x01' * 32
    k = b'\x00' * 32
    priv = encode_privkey(priv, 'bin')
    msghash = encode(hash_to_int(msghash), 256, 32)
    k = hmac.new(k, v+b'\x00'+priv+msghash, hashlib.sha256).digest()
    v = hmac.new(k, v, hashlib.sha256).digest()
    k = hmac.new(k, v+b'\x01'+priv+msghash, hashlib.sha256).digest()
    v = hmac.new(k, v, hashlib.sha256).digest()
    return decode(hmac.new(k, v, hashlib.sha256).digest(), 256)


def ecdsa_raw_sign(msghash, priv):

    z = hash_to_int(msghash)
    k = deterministic_generate_k(msghash, priv)

    r, y = fast_multiply(G, k)
    s = inv(k, N) * (z + r*decode_privkey(priv)) % N

    v, r, s = 27+((y % 2) ^ (0 if s * 2 < N else 1)), r, s if s * 2 < N else N - s
    if 'compressed' in get_privkey_format(priv):
        v += 4
    return v, r, s


def ecdsa_sign(msg, priv, coin):
    v, r, s = ecdsa_raw_sign(electrum_sig_hash(msg), priv)
    sig = encode_sig(v, r, s)
    assert ecdsa_verify(msg, sig, 
        privtopub(priv), coin), "Bad Sig!\t %s\nv = %d\n,r = %d\ns = %d" % (sig, v, r, s)
    return sig


def ecdsa_raw_verify(msghash, vrs, pub):
    v, r, s = vrs

    w = inv(s, N)
    z = hash_to_int(msghash)

    u1, u2 = z*w % N, r*w % N
    x, y = fast_add(fast_multiply(G, u1), fast_multiply(decode_pubkey(pub), u2))
    return bool(r == x and (r % N) and (s % N))


# For BitcoinCore, (msg = addr or msg = "") be default
def ecdsa_verify_addr(msg, sig, addr, coin):
    assert coin.is_address(addr)
    Q = ecdsa_recover(msg, sig)
    magic = get_version_byte(addr)
    return (addr == coin.pubtoaddr(Q, int(magic))) or (addr == coin.pubtoaddr(compress(Q), int(magic)))


def ecdsa_verify(msg, sig, pub, coin):
    if coin.is_address(pub):
        return ecdsa_verify_addr(msg, sig, pub, coin)
    return ecdsa_raw_verify(electrum_sig_hash(msg), decode_sig(sig), pub)


def ecdsa_raw_recover(msghash, vrs):
    v, r, s = vrs
    x = r
    xcubedaxb = (x*x*x+A*x+B) % P
    beta = pow(xcubedaxb, (P+1)//4, P)
    y = beta if v % 2 ^ beta % 2 else (P - beta)
    # If xcubedaxb is not a quadratic residue, then r cannot be the x coord
    # for a point on the curve, and so the sig is invalid
    if (xcubedaxb - y*y) % P != 0 or not (r % N) or not (s % N):
        return False
    z = hash_to_int(msghash)
    Gz = jacobian_multiply((Gx, Gy, 1), (N - z) % N)
    XY = jacobian_multiply((x, y, 1), s)
    Qr = jacobian_add(Gz, XY)
    Q = jacobian_multiply(Qr, inv(r, N))
    Q = from_jacobian(Q)

    # if ecdsa_raw_verify(msghash, vrs, Q):
    return Q
    # return False


def ecdsa_recover(msg, sig):
    v,r,s = decode_sig(sig)
    Q = ecdsa_raw_recover(electrum_sig_hash(msg), (v,r,s))
    return encode_pubkey(Q, 'hex_compressed') if v >= 31 else encode_pubkey(Q, 'hex')



# add/subtract 
def add(p1,p2):
    if is_privkey(p1):
        return add_privkeys(p1, p2)
    else:
        return add_pubkeys(p1, p2)

def subtract(p1,p2):
    if is_privkey(p1):
        return subtract_privkeys(p1, p2)
    else:
        return subtract_pubkeys(p1, p2)

hash160Low = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
hash160High = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'

def magicbyte_to_prefix(magicbyte):
    first = bin_to_b58check(hash160Low, magicbyte=magicbyte)[0]
    last = bin_to_b58check(hash160High, magicbyte=magicbyte)[0]
    if first == last:
        return (first,)
    return (first, last)



from .main import *
import hmac
import hashlib

# Electrum wallets


def electrum_stretch(seed):
    return slowsha(seed)

# Accepts seed or stretched seed, returns master public key


def electrum_mpk(seed):
    if len(seed) == 32:
        seed = electrum_stretch(seed)
    return privkey_to_pubkey(seed)[2:]

# Accepts (seed or stretched seed), index and secondary index
# (conventionally 0 for ordinary addresses, 1 for change) , returns privkey


def electrum_privkey(seed, n, for_change=0):
    if len(seed) == 32:
        seed = electrum_stretch(seed)
    mpk = electrum_mpk(seed)
    offset = dbl_sha256(from_int_representation_to_bytes(n)+b':'+from_int_representation_to_bytes(for_change)+b':'+binascii.unhexlify(mpk))
    return add_privkeys(seed, offset)

# Accepts (seed or stretched seed or master pubkey), index and secondary index
# (conventionally 0 for ordinary addresses, 1 for change) , returns pubkey


def electrum_pubkey(masterkey, n, for_change=0):
    if len(masterkey) == 32:
        mpk = electrum_mpk(electrum_stretch(masterkey))
    elif len(masterkey) == 64:
        mpk = electrum_mpk(masterkey)
    else:
        mpk = masterkey
    bin_mpk = encode_pubkey(mpk, 'bin_electrum')
    offset = bin_dbl_sha256(from_int_representation_to_bytes(n)+b':'+from_int_representation_to_bytes(for_change)+b':'+bin_mpk)
    return add_pubkeys('04'+mpk, privtopub(offset))

# seed/stretched seed/pubkey -> address (convenience method)


def electrum_address(masterkey, n, for_change=0, magicbyte=0):
    return pubkey_to_address(electrum_pubkey(masterkey, n, for_change), magicbyte)

# Given a master public key, a private key from that wallet and its index,
# cracks the secret exponent which can be used to generate all other private
# keys in the wallet


def crack_electrum_wallet(mpk, pk, n, for_change=0):
    bin_mpk = encode_pubkey(mpk, 'bin_electrum')
    offset = dbl_sha256(str(n)+':'+str(for_change)+':'+bin_mpk)
    return subtract_privkeys(pk, offset)

# Below code ASSUMES binary inputs and compressed pubkeys
MAINNET_PRIVATE = b'\x04\x88\xAD\xE4'
MAINNET_PUBLIC = b'\x04\x88\xB2\x1E'
TESTNET_PRIVATE = b'\x04\x35\x83\x94'
TESTNET_PUBLIC = b'\x04\x35\x87\xCF'
PRIVATE = [MAINNET_PRIVATE, TESTNET_PRIVATE]
PUBLIC = [MAINNET_PUBLIC, TESTNET_PUBLIC]
DEFAULT = (MAINNET_PRIVATE, MAINNET_PUBLIC)

# BIP32 child key derivation


def raw_bip32_ckd(rawtuple, i, prefixes=DEFAULT):
    vbytes, depth, fingerprint, oldi, chaincode, key = rawtuple
    i = int(i)

    private = vbytes == prefixes[0]

    if private:
        priv = key
        pub = privtopub(key)
    else:
        priv = None
        pub = key

    if i >= 2**31:
        if not priv:
            raise Exception("Can't do private derivation on public key!")
        I = hmac.new(chaincode, b'\x00'+priv[:32]+encode(i, 256, 4), hashlib.sha512).digest()
    else:
        I = hmac.new(chaincode, pub+encode(i, 256, 4), hashlib.sha512).digest()
    if private:
        newkey = add_privkeys(I[:32]+B'\x01', priv)
        fingerprint = bin_hash160(privtopub(key))[:4]
    else:
        newkey = add_pubkeys(compress(privtopub(I[:32])), key)
        fingerprint = bin_hash160(key)[:4]

    return (vbytes, depth + 1, fingerprint, i, I[32:], newkey)


def bip32_serialize(rawtuple, prefixes=DEFAULT):
    vbytes, depth, fingerprint, i, chaincode, key = rawtuple
    i = encode(i, 256, 4)
    chaincode = encode(hash_to_int(chaincode), 256, 32)
    keydata = b'\x00'+key[:-1] if vbytes == prefixes[0] else key
    bindata = vbytes + from_int_to_byte(depth % 256) + fingerprint + i + chaincode + keydata
    return changebase(bindata+bin_dbl_sha256(bindata)[:4], 256, 58)


def bip32_deserialize(data, prefixes=DEFAULT):
    dbin = changebase(data, 58, 256)
    if bin_dbl_sha256(dbin[:-4])[:4] != dbin[-4:]:
        raise Exception("Invalid checksum")
    vbytes = dbin[0:4]
    depth = from_byte_to_int(dbin[4])
    fingerprint = dbin[5:9]
    i = decode(dbin[9:13], 256)
    chaincode = dbin[13:45]
    key = dbin[46:78]+b'\x01' if vbytes == prefixes[0] else dbin[45:78]
    return (vbytes, depth, fingerprint, i, chaincode, key)


def is_xprv(text, prefixes=DEFAULT):
    vbytes, depth, fingerprint, i, chaincode, key = bip32_deserialize(text, prefixes)
    return vbytes == prefixes[0]


def is_xpub(text, prefixes=DEFAULT):
    vbytes, depth, fingerprint, i, chaincode, key = bip32_deserialize(text, prefixes)
    return vbytes == prefixes[1]


def raw_bip32_privtopub(rawtuple, prefixes=DEFAULT):
    vbytes, depth, fingerprint, i, chaincode, key = rawtuple
    newvbytes = prefixes[1]
    return (newvbytes, depth, fingerprint, i, chaincode, privtopub(key))


def bip32_privtopub(data, prefixes=DEFAULT):
    return bip32_serialize(raw_bip32_privtopub(bip32_deserialize(data, prefixes), prefixes), prefixes)


def bip32_ckd(key, path, prefixes=DEFAULT, public=False):
    if isinstance(path, (list, tuple)):
        pathlist = map(str, path)
    else:
        path = str(path)
        pathlist = parse_bip32_path(path)
    for i, p in enumerate(pathlist):
        key = bip32_serialize(raw_bip32_ckd(bip32_deserialize(key, prefixes), p, prefixes), prefixes)
    return key if not public else bip32_privtopub(key)

def bip32_master_key(seed, prefixes=DEFAULT):
    I = hmac.new(
            from_string_to_bytes("Bitcoin seed"), 
            from_string_to_bytes(seed), 
            hashlib.sha512
        ).digest()
    return bip32_serialize((prefixes[0], 0, b'\x00'*4, 0, I[32:], I[:32]+b'\x01'), prefixes)


def bip32_bin_extract_key(data, prefixes=DEFAULT):
    return bip32_deserialize(data, prefixes)[-1]


def bip32_extract_key(data, prefixes=DEFAULT):
    return safe_hexlify(bip32_deserialize(data, prefixes)[-1])


def bip32_derive_key(key, path, prefixes=DEFAULT, **kwargs):
    return bip32_extract_key(bip32_ckd(key, path, prefixes, **kwargs), prefixes)

# Exploits the same vulnerability as above in Electrum wallets
# Takes a BIP32 pubkey and one of the child privkeys of its corresponding
# privkey and returns the BIP32 privkey associated with that pubkey


def raw_crack_bip32_privkey(parent_pub, priv, prefixes=DEFAULT):
    vbytes, depth, fingerprint, i, chaincode, key = priv
    pvbytes, pdepth, pfingerprint, pi, pchaincode, pkey = parent_pub
    i = int(i)

    if i >= 2**31:
        raise Exception("Can't crack private derivation!")

    I = hmac.new(pchaincode, pkey+encode(i, 256, 4), hashlib.sha512).digest()

    pprivkey = subtract_privkeys(key, I[:32]+b'\x01')

    newvbytes = prefixes[0]
    return (newvbytes, pdepth, pfingerprint, pi, pchaincode, pprivkey)


def crack_bip32_privkey(parent_pub, priv, prefixes=DEFAULT):
    dsppub = bip32_deserialize(parent_pub, prefixes)
    dspriv = bip32_deserialize(priv, prefixes)
    return bip32_serialize(raw_crack_bip32_privkey(dsppub, dspriv, prefixes), prefixes)


def coinvault_pub_to_bip32(*args, prefixes=DEFAULT):
    if len(args) == 1:
        args = args[0].split(' ')
    vals = map(int, args[34:])
    I1 = ''.join(map(chr, vals[:33]))
    I2 = ''.join(map(chr, vals[35:67]))
    return bip32_serialize((prefixes[1], 0, b'\x00'*4, 0, I2, I1))


def coinvault_priv_to_bip32(*args, prefixes=DEFAULT):
    if len(args) == 1:
        args = args[0].split(' ')
    vals = map(int, args[34:])
    I2 = ''.join(map(chr, vals[35:67]))
    I3 = ''.join(map(chr, vals[72:104]))
    return bip32_serialize((prefixes[0], 0, b'\x00'*4, 0, I2, I3+b'\x01'))


def bip32_descend(*args, prefixes=DEFAULT):
    """Descend masterkey and return privkey"""
    if len(args) == 2 and isinstance(args[1], list):
        key, path = args
    elif len(args) == 2 and isinstance(args[1], string_types):
        key = args[0]
        path = map(int, str(args[1]).lstrip("mM/").split('/'))
    elif len(args):
        key, path = args[0], map(int, args[1:])
    for p in path:
        key = bip32_ckd(key, p, prefixes)
    return bip32_extract_key(key, prefixes)

def parse_bip32_path(path):
    """Takes bip32 path, "m/0'/2H" or "m/0H/1/2H/2/1000000000.pub", returns list of ints """
    path = path.lstrip("m/").rstrip(".pub")
    if not path:
        return []
    #elif path.endswith("/"):       incorrect for electrum segwit
    #    path += "0"
    patharr = []
    for v in path.split('/'):
        if not v: 
            continue
        elif v[-1] in ("'H"):  # hardened path
            v = int(v[:-1]) | 0x80000000
        else:                  # non-hardened path
            v = int(v) & 0x7fffffff
        patharr.append(v)
    return patharr



from . import main

# Shared secrets and uncovering pay keys


def shared_secret_sender(scan_pubkey, ephem_privkey):
    shared_point = main.multiply(scan_pubkey, ephem_privkey)
    shared_secret = main.sha256(main.encode_pubkey(shared_point, 'bin_compressed'))
    return shared_secret


def shared_secret_receiver(ephem_pubkey, scan_privkey):
    shared_point = main.multiply(ephem_pubkey, scan_privkey)
    shared_secret = main.sha256(main.encode_pubkey(shared_point, 'bin_compressed'))
    return shared_secret


def uncover_pay_pubkey_sender(scan_pubkey, spend_pubkey, ephem_privkey):
    shared_secret = shared_secret_sender(scan_pubkey, ephem_privkey)
    return main.add_pubkeys(spend_pubkey, main.privtopub(shared_secret))


def uncover_pay_pubkey_receiver(scan_privkey, spend_pubkey, ephem_pubkey):
    shared_secret = shared_secret_receiver(ephem_pubkey, scan_privkey)
    return main.add_pubkeys(spend_pubkey, main.privtopub(shared_secret))


def uncover_pay_privkey(scan_privkey, spend_privkey, ephem_pubkey):
    shared_secret = shared_secret_receiver(ephem_pubkey, scan_privkey)
    return main.add_privkeys(spend_privkey, shared_secret)

# Address encoding

# Functions for basic stealth addresses,
# i.e. one scan key, one spend key, no prefix


def pubkeys_to_basic_stealth_address(scan_pubkey, spend_pubkey, magic_byte=42):
    # magic_byte = 42 for mainnet, 43 for testnet.
    hex_scankey = main.encode_pubkey(scan_pubkey, 'hex_compressed')
    hex_spendkey = main.encode_pubkey(spend_pubkey, 'hex_compressed')
    hex_data = '00{0:066x}01{1:066x}0100'.format(int(hex_scankey, 16), int(hex_spendkey, 16))
    addr = main.hex_to_b58check(hex_data, magic_byte)
    return addr


def basic_stealth_address_to_pubkeys(stealth_address):
    hex_data = main.b58check_to_hex(stealth_address)
    if len(hex_data) != 140:
        raise Exception('Stealth address is not of basic type (one scan key, one spend key, no prefix)')

    scan_pubkey = hex_data[2:68]
    spend_pubkey = hex_data[70:136]
    return scan_pubkey, spend_pubkey

# Sending stealth payments


def mk_stealth_metadata_script(ephem_pubkey, nonce):
    op_return = '6a'
    msg_size = '26'
    version = '06'
    return op_return + msg_size + version + '{0:08x}'.format(nonce) + main.encode_pubkey(ephem_pubkey, 'hex_compressed')


def mk_stealth_tx_outputs(stealth_addr, value, ephem_privkey, nonce, network='btc'):

    scan_pubkey, spend_pubkey = basic_stealth_address_to_pubkeys(stealth_addr)

    if network == 'btc':
        btc_magic_byte = 42
        if stealth_addr != pubkeys_to_basic_stealth_address(scan_pubkey, spend_pubkey, btc_magic_byte):
            raise Exception('Invalid btc mainnet stealth address: ' + stealth_addr)
        magic_byte_addr = 0

    elif network == 'testnet':
        testnet_magic_byte = 43
        if stealth_addr != pubkeys_to_basic_stealth_address(scan_pubkey, spend_pubkey, testnet_magic_byte):
            raise Exception('Invalid testnet stealth address: ' + stealth_addr)
        magic_byte_addr = 111

    ephem_pubkey = main.privkey_to_pubkey(ephem_privkey)
    output0 = {'script': mk_stealth_metadata_script(ephem_pubkey, nonce),
               'value': 0}

    pay_pubkey = uncover_pay_pubkey_sender(scan_pubkey, spend_pubkey, ephem_privkey)
    pay_addr = main.pubkey_to_address(pay_pubkey, magic_byte_addr)
    output1 = {'address': pay_addr,
               'value': value}

    return [output0, output1]

# Receiving stealth payments


def ephem_pubkey_from_tx_script(stealth_tx_script):
    if len(stealth_tx_script) != 80:
        raise Exception('Wrong format for stealth tx output')
    return stealth_tx_script[14:]



import re
import requests

#Docs: https://chain.so/api

base_url = "https://chain.so/api/v2/"
sendtx_url = base_url + "send_tx/%s"
address_url = base_url + "address/%s/%s"
utxo_url = base_url + "get_tx_unspent/%s/%s"
tx_url = base_url + "get_tx/%s/%s"
tx_details_url = base_url + "tx/%s/%s"
tx_inputs_url = base_url + "get_tx_inputs/%s/%s"
network_url = base_url + "get_info/%s"
block_url = base_url + "block/%s/%s"

def unspent(addr, coin_symbol="LTC"):
    url = utxo_url % (coin_symbol, addr)
    response = requests.get(url)
    try:
        result = response.json()
        if 'data' in result.keys() and 'txs' in result['data'].keys():
            txs = response.json()['data']['txs']
            for i, tx in enumerate(txs):
                txs[i] = {
                    'output': "%s:%s" % (tx['txid'], tx['output_no']),
                    'value': int(tx['value'].replace('.', '')),
                }
            return txs
        else:
            raise Exception(response.text)
    except ValueError:
        raise Exception("Unable to decode JSON from result: %s" % response.text)

def fetchtx(txhash, coin_symbol="LTC"):
    url = tx_url % (coin_symbol, txhash)
    response = requests.get(url)
    result = response.json()
    return result['data']

def gettxdetails(txhash, coin_symbol="LTC"):
    url = tx_details_url % (coin_symbol, txhash)
    response = requests.get(url)
    result = response.json()
    return result['data']

def txinputs(txhash, coin_symbol="LTC"):
    url = tx_inputs_url % (coin_symbol, txhash)
    response = requests.get(url)
    result = response.json()
    inputs = result['data']['inputs']
    unspents = [{'output': (i['from_output']['txid'] + ":" + str(i['from_output']['output_no'])), 'value': int(float(i['value']) * 100000000)} for i in inputs]
    return unspents

def pushtx(tx, coin_symbol="LTC"):
    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')
    url = sendtx_url % coin_symbol
    response = requests.post(url, {'tx_hex': tx})
    return response.json()

def history(addr, coin_symbol="LTC"):
    url = address_url % (coin_symbol, addr)
    response = requests.get(url)
    return response.json()

def block_height(txhash, coin_symbol="LTC"):
    tx = gettxdetails(txhash,coin_symbol=coin_symbol)
    return tx['block_no']

def block_info(height, coin_symbol="LTC"):
    url = block_url % (coin_symbol, height)
    response = requests.get(url)
    data = response.json()['data']
    return {
        'version': data['version'],
        'hash': data['blockhash'],
        'prevhash': data['previous_blockhash'],
        'timestamp': data['time'],
        'merkle_root': data['merkleroot'],
        'bits': data['bits'],
        'nonce': data['nonce'],
        'tx_hashes': [t['txid'] for t in data['txs']]
    }


def current_block_height(coin_symbol="LTC"):
    url = network_url % coin_symbol
    response = requests.get(url)
    return response.json()['data']['blocks']



# -*- coding: utf-8 -*-
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2011 thomasv@gitorious with changes by pycryptools developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import hashlib
import hmac
from .main import *
from .py2specials import *
from .py3specials import *
from . import constants as version

# Version numbers for BIP32 extended keys
# standard: xprv, xpub
# segwit in p2sh: yprv, ypub
# native segwit: zprv, zpub
XPRV_HEADERS = {
    'standard': 0x0488ade4,
    'p2wpkh-p2sh': 0x049d7878,
    'p2wsh-p2sh': 0x295b005,
    'p2wpkh': 0x4b2430c,
    'p2wsh': 0x2aa7a99
}
XPUB_HEADERS = {
    'standard': 0x0488b21e,
    'p2wpkh-p2sh': 0x049d7cb2,
    'p2wsh-p2sh': 0x295b43f,
    'p2wpkh': 0x4b24746,
    'p2wsh': 0x2aa7ed3
}

bh2u = safe_hexlify
hfu = binascii.hexlify
bfh = safe_from_hex
hmac_sha_512 = lambda x, y: hmac.new(x, y, hashlib.sha512).digest()

def rev_hex(s):
    return bh2u(bfh(s)[::-1])

def int_to_hex(i, length=1):
    assert isinstance(i, int)
    s = hex(i)[2:].rstrip('L')
    s = "0"*(2*length - len(s)) + s
    return rev_hex(s)

class InvalidPassword(Exception):
    def __str__(self):
        return "Incorrect password"

try:
    from Cryptodome.Cipher import AES
except:
    AES = None

class InvalidPasswordException(Exception):
    pass

class InvalidPadding(Exception):
    pass

def assert_bytes(*args):
    """
    porting helper, assert args type
    """
    try:
        for x in args:
            assert isinstance(x, (bytes, bytearray))
    except:
        print('assert bytes failed', list(map(type, args)))
        raise

def append_PKCS7_padding(data):
    assert_bytes(data)
    padlen = 16 - (len(data) % 16)
    return data + bytes([padlen]) * padlen


def strip_PKCS7_padding(data):
    assert_bytes(data)
    if len(data) % 16 != 0 or len(data) == 0:
        raise InvalidPadding("invalid length")
    padlen = data[-1]
    if padlen > 16:
        raise InvalidPadding("invalid padding byte (large)")
    for i in data[-padlen:]:
        if i != padlen:
            raise InvalidPadding("invalid padding byte (inconsistent)")
    return data[0:-padlen]

def aes_encrypt_with_iv(key, iv, data):
    assert_bytes(key, iv, data)
    data = append_PKCS7_padding(data)
    if AES:
        e = AES.new(key, AES.MODE_CBC, iv).encrypt(data)
    else:
        aes_cbc = pyaes.AESModeOfOperationCBC(key, iv=iv)
        aes = pyaes.Encrypter(aes_cbc, padding=pyaes.PADDING_NONE)
        e = aes.feed(data) + aes.feed()  # empty aes.feed() flushes buffer
    return e


def aes_decrypt_with_iv(key, iv, data):
    assert_bytes(key, iv, data)
    if AES:
        cipher = AES.new(key, AES.MODE_CBC, iv)
        data = cipher.decrypt(data)
    else:
        aes_cbc = pyaes.AESModeOfOperationCBC(key, iv=iv)
        aes = pyaes.Decrypter(aes_cbc, padding=pyaes.PADDING_NONE)
        data = aes.feed(data) + aes.feed()  # empty aes.feed() flushes buffer
    try:
        return strip_PKCS7_padding(data)
    except InvalidPadding:
        raise InvalidPassword()

def EncodeAES(secret, s):
    assert_bytes(s)
    iv = bytes(os.urandom(16))
    ct = aes_encrypt_with_iv(secret, iv, s)
    e = iv + ct
    return base64.b64encode(e)

def DecodeAES(secret, e):
    e = bytes(base64.b64decode(e))
    iv, e = e[:16], e[16:]
    s = aes_decrypt_with_iv(secret, iv, e)
    return s

def pw_encode(s, password):
    if password:
        secret = Hash(password)
        return EncodeAES(secret, to_bytes(s, "utf8")).decode('utf8')
    else:
        return s

def pw_decode(s, password):
    if password is not None:
        secret = Hash(password)
        try:
            d = to_string(DecodeAES(secret, s), "utf8")
        except Exception:
            raise InvalidPassword()
        return d
    else:
        return s

def is_new_seed(x, prefix=version.SEED_PREFIX):
    from . import mnemonic
    x = mnemonic.normalize_text(x)
    s = bh2u(hmac_sha_512(b"Seed version", x.encode('utf8')))
    return s.startswith(prefix)

def seed_type(x):
    if is_new_seed(x):
        return 'standard'
    elif is_new_seed(x, version.SEED_PREFIX_SW):
        return 'segwit'
    elif is_new_seed(x, version.SEED_PREFIX_2FA):
        return '2fa'
    return ''

is_seed = lambda x: bool(seed_type(x))

SCRIPT_TYPES = {
    'p2pkh':0,
    'p2wpkh':1,
    'p2wpkh-p2sh':2,
    'p2sh':5,
    'p2wsh':6,
    'p2wsh-p2sh':7
}

__b58chars = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
assert len(__b58chars) == 58

__b43chars = b'0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ$*+-./:'
assert len(__b43chars) == 43

def inv_dict(d):
    return {v: k for k, v in d.items()}

def is_minikey(text):
    # Minikeys are typically 22 or 30 characters, but this routine
    # permits any length of 20 or more provided the minikey is valid.
    # A valid minikey must begin with an 'S', be in base58, and when
    # suffixed with '?' have its SHA256 hash begin with a zero byte.
    # They are widely used in Casascius physical bitcoins.
    return (len(text) >= 20 and text[0] == 'S'
            and all(ord(c) in __b58chars for c in text)
            and sha256(text + '?')[0] == 0x00)

def minikey_to_private_key(text):
    return sha256(text)


###################################### BIP32 ##############################

BIP32_PRIME = 0x80000000


def get_pubkeys_from_secret(secret):
    # public key
    pubkey = compress(privtopub(secret))
    return pubkey, True

def xprv_header(xtype):
    return bfh("%08x" % XPRV_HEADERS[xtype])


def xpub_header(xtype):
    return bfh("%08x" % XPUB_HEADERS[xtype])


#!/usr/bin/python
# -*- coding: utf-8 -*-
s = """
abaisser
abandon
abdiquer
abeille
abolir
aborder
aboutir
aboyer
abrasif
abreuver
abriter
abroger
abrupt
absence
absolu
absurde
abusif
abyssal
académie
acajou
acarien
accabler
accepter
acclamer
accolade
accroche
accuser
acerbe
achat
acheter
aciduler
acier
acompte
acquérir
acronyme
acteur
actif
actuel
adepte
adéquat
adhésif
adjectif
adjuger
admettre
admirer
adopter
adorer
adoucir
adresse
adroit
adulte
adverbe
aérer
aéronef
affaire
affecter
affiche
affreux
affubler
agacer
agencer
agile
agiter
agrafer
agréable
agrume
aider
aiguille
ailier
aimable
aisance
ajouter
ajuster
alarmer
alchimie
alerte
algèbre
algue
aliéner
aliment
alléger
alliage
allouer
allumer
alourdir
alpaga
altesse
alvéole
amateur
ambigu
ambre
aménager
amertume
amidon
amiral
amorcer
amour
amovible
amphibie
ampleur
amusant
analyse
anaphore
anarchie
anatomie
ancien
anéantir
angle
angoisse
anguleux
animal
annexer
annonce
annuel
anodin
anomalie
anonyme
anormal
antenne
antidote
anxieux
apaiser
apéritif
aplanir
apologie
appareil
appeler
apporter
appuyer
aquarium
aqueduc
arbitre
arbuste
ardeur
ardoise
argent
arlequin
armature
armement
armoire
armure
arpenter
arracher
arriver
arroser
arsenic
artériel
article
aspect
asphalte
aspirer
assaut
asservir
assiette
associer
assurer
asticot
astre
astuce
atelier
atome
atrium
atroce
attaque
attentif
attirer
attraper
aubaine
auberge
audace
audible
augurer
aurore
automne
autruche
avaler
avancer
avarice
avenir
averse
aveugle
aviateur
avide
avion
aviser
avoine
avouer
avril
axial
axiome
badge
bafouer
bagage
baguette
baignade
balancer
balcon
baleine
balisage
bambin
bancaire
bandage
banlieue
bannière
banquier
barbier
baril
baron
barque
barrage
bassin
bastion
bataille
bateau
batterie
baudrier
bavarder
belette
bélier
belote
bénéfice
berceau
berger
berline
bermuda
besace
besogne
bétail
beurre
biberon
bicycle
bidule
bijou
bilan
bilingue
billard
binaire
biologie
biopsie
biotype
biscuit
bison
bistouri
bitume
bizarre
blafard
blague
blanchir
blessant
blinder
blond
bloquer
blouson
bobard
bobine
boire
boiser
bolide
bonbon
bondir
bonheur
bonifier
bonus
bordure
borne
botte
boucle
boueux
bougie
boulon
bouquin
bourse
boussole
boutique
boxeur
branche
brasier
brave
brebis
brèche
breuvage
bricoler
brigade
brillant
brioche
brique
brochure
broder
bronzer
brousse
broyeur
brume
brusque
brutal
bruyant
buffle
buisson
bulletin
bureau
burin
bustier
butiner
butoir
buvable
buvette
cabanon
cabine
cachette
cadeau
cadre
caféine
caillou
caisson
calculer
calepin
calibre
calmer
calomnie
calvaire
camarade
caméra
camion
campagne
canal
caneton
canon
cantine
canular
capable
caporal
caprice
capsule
capter
capuche
carabine
carbone
caresser
caribou
carnage
carotte
carreau
carton
cascade
casier
casque
cassure
causer
caution
cavalier
caverne
caviar
cédille
ceinture
céleste
cellule
cendrier
censurer
central
cercle
cérébral
cerise
cerner
cerveau
cesser
chagrin
chaise
chaleur
chambre
chance
chapitre
charbon
chasseur
chaton
chausson
chavirer
chemise
chenille
chéquier
chercher
cheval
chien
chiffre
chignon
chimère
chiot
chlorure
chocolat
choisir
chose
chouette
chrome
chute
cigare
cigogne
cimenter
cinéma
cintrer
circuler
cirer
cirque
citerne
citoyen
citron
civil
clairon
clameur
claquer
classe
clavier
client
cligner
climat
clivage
cloche
clonage
cloporte
cobalt
cobra
cocasse
cocotier
coder
codifier
coffre
cogner
cohésion
coiffer
coincer
colère
colibri
colline
colmater
colonel
combat
comédie
commande
compact
concert
conduire
confier
congeler
connoter
consonne
contact
convexe
copain
copie
corail
corbeau
cordage
corniche
corpus
correct
cortège
cosmique
costume
coton
coude
coupure
courage
couteau
couvrir
coyote
crabe
crainte
cravate
crayon
créature
créditer
crémeux
creuser
crevette
cribler
crier
cristal
critère
croire
croquer
crotale
crucial
cruel
crypter
cubique
cueillir
cuillère
cuisine
cuivre
culminer
cultiver
cumuler
cupide
curatif
curseur
cyanure
cycle
cylindre
cynique
daigner
damier
danger
danseur
dauphin
débattre
débiter
déborder
débrider
débutant
décaler
décembre
déchirer
décider
déclarer
décorer
décrire
décupler
dédale
déductif
déesse
défensif
défiler
défrayer
dégager
dégivrer
déglutir
dégrafer
déjeuner
délice
déloger
demander
demeurer
démolir
dénicher
dénouer
dentelle
dénuder
départ
dépenser
déphaser
déplacer
déposer
déranger
dérober
désastre
descente
désert
désigner
désobéir
dessiner
destrier
détacher
détester
détourer
détresse
devancer
devenir
deviner
devoir
diable
dialogue
diamant
dicter
différer
digérer
digital
digne
diluer
dimanche
diminuer
dioxyde
directif
diriger
discuter
disposer
dissiper
distance
divertir
diviser
docile
docteur
dogme
doigt
domaine
domicile
dompter
donateur
donjon
donner
dopamine
dortoir
dorure
dosage
doseur
dossier
dotation
douanier
double
douceur
douter
doyen
dragon
draper
dresser
dribbler
droiture
duperie
duplexe
durable
durcir
dynastie
éblouir
écarter
écharpe
échelle
éclairer
éclipse
éclore
écluse
école
économie
écorce
écouter
écraser
écrémer
écrivain
écrou
écume
écureuil
édifier
éduquer
effacer
effectif
effigie
effort
effrayer
effusion
égaliser
égarer
éjecter
élaborer
élargir
électron
élégant
éléphant
élève
éligible
élitisme
éloge
élucider
éluder
emballer
embellir
embryon
émeraude
émission
emmener
émotion
émouvoir
empereur
employer
emporter
emprise
émulsion
encadrer
enchère
enclave
encoche
endiguer
endosser
endroit
enduire
énergie
enfance
enfermer
enfouir
engager
engin
englober
énigme
enjamber
enjeu
enlever
ennemi
ennuyeux
enrichir
enrobage
enseigne
entasser
entendre
entier
entourer
entraver
énumérer
envahir
enviable
envoyer
enzyme
éolien
épaissir
épargne
épatant
épaule
épicerie
épidémie
épier
épilogue
épine
épisode
épitaphe
époque
épreuve
éprouver
épuisant
équerre
équipe
ériger
érosion
erreur
éruption
escalier
espadon
espèce
espiègle
espoir
esprit
esquiver
essayer
essence
essieu
essorer
estime
estomac
estrade
étagère
étaler
étanche
étatique
éteindre
étendoir
éternel
éthanol
éthique
ethnie
étirer
étoffer
étoile
étonnant
étourdir
étrange
étroit
étude
euphorie
évaluer
évasion
éventail
évidence
éviter
évolutif
évoquer
exact
exagérer
exaucer
exceller
excitant
exclusif
excuse
exécuter
exemple
exercer
exhaler
exhorter
exigence
exiler
exister
exotique
expédier
explorer
exposer
exprimer
exquis
extensif
extraire
exulter
fable
fabuleux
facette
facile
facture
faiblir
falaise
fameux
famille
farceur
farfelu
farine
farouche
fasciner
fatal
fatigue
faucon
fautif
faveur
favori
fébrile
féconder
fédérer
félin
femme
fémur
fendoir
féodal
fermer
féroce
ferveur
festival
feuille
feutre
février
fiasco
ficeler
fictif
fidèle
figure
filature
filetage
filière
filleul
filmer
filou
filtrer
financer
finir
fiole
firme
fissure
fixer
flairer
flamme
flasque
flatteur
fléau
flèche
fleur
flexion
flocon
flore
fluctuer
fluide
fluvial
folie
fonderie
fongible
fontaine
forcer
forgeron
formuler
fortune
fossile
foudre
fougère
fouiller
foulure
fourmi
fragile
fraise
franchir
frapper
frayeur
frégate
freiner
frelon
frémir
frénésie
frère
friable
friction
frisson
frivole
froid
fromage
frontal
frotter
fruit
fugitif
fuite
fureur
furieux
furtif
fusion
futur
gagner
galaxie
galerie
gambader
garantir
gardien
garnir
garrigue
gazelle
gazon
géant
gélatine
gélule
gendarme
général
génie
genou
gentil
géologie
géomètre
géranium
germe
gestuel
geyser
gibier
gicler
girafe
givre
glace
glaive
glisser
globe
gloire
glorieux
golfeur
gomme
gonfler
gorge
gorille
goudron
gouffre
goulot
goupille
gourmand
goutte
graduel
graffiti
graine
grand
grappin
gratuit
gravir
grenat
griffure
griller
grimper
grogner
gronder
grotte
groupe
gruger
grutier
gruyère
guépard
guerrier
guide
guimauve
guitare
gustatif
gymnaste
gyrostat
habitude
hachoir
halte
hameau
hangar
hanneton
haricot
harmonie
harpon
hasard
hélium
hématome
herbe
hérisson
hermine
héron
hésiter
heureux
hiberner
hibou
hilarant
histoire
hiver
homard
hommage
homogène
honneur
honorer
honteux
horde
horizon
horloge
hormone
horrible
houleux
housse
hublot
huileux
humain
humble
humide
humour
hurler
hydromel
hygiène
hymne
hypnose
idylle
ignorer
iguane
illicite
illusion
image
imbiber
imiter
immense
immobile
immuable
impact
impérial
implorer
imposer
imprimer
imputer
incarner
incendie
incident
incliner
incolore
indexer
indice
inductif
inédit
ineptie
inexact
infini
infliger
informer
infusion
ingérer
inhaler
inhiber
injecter
injure
innocent
inoculer
inonder
inscrire
insecte
insigne
insolite
inspirer
instinct
insulter
intact
intense
intime
intrigue
intuitif
inutile
invasion
inventer
inviter
invoquer
ironique
irradier
irréel
irriter
isoler
ivoire
ivresse
jaguar
jaillir
jambe
janvier
jardin
jauger
jaune
javelot
jetable
jeton
jeudi
jeunesse
joindre
joncher
jongler
joueur
jouissif
journal
jovial
joyau
joyeux
jubiler
jugement
junior
jupon
juriste
justice
juteux
juvénile
kayak
kimono
kiosque
label
labial
labourer
lacérer
lactose
lagune
laine
laisser
laitier
lambeau
lamelle
lampe
lanceur
langage
lanterne
lapin
largeur
larme
laurier
lavabo
lavoir
lecture
légal
léger
légume
lessive
lettre
levier
lexique
lézard
liasse
libérer
libre
licence
licorne
liège
lièvre
ligature
ligoter
ligue
limer
limite
limonade
limpide
linéaire
lingot
lionceau
liquide
lisière
lister
lithium
litige
littoral
livreur
logique
lointain
loisir
lombric
loterie
louer
lourd
loutre
louve
loyal
lubie
lucide
lucratif
lueur
lugubre
luisant
lumière
lunaire
lundi
luron
lutter
luxueux
machine
magasin
magenta
magique
maigre
maillon
maintien
mairie
maison
majorer
malaxer
maléfice
malheur
malice
mallette
mammouth
mandater
maniable
manquant
manteau
manuel
marathon
marbre
marchand
mardi
maritime
marqueur
marron
marteler
mascotte
massif
matériel
matière
matraque
maudire
maussade
mauve
maximal
méchant
méconnu
médaille
médecin
méditer
méduse
meilleur
mélange
mélodie
membre
mémoire
menacer
mener
menhir
mensonge
mentor
mercredi
mérite
merle
messager
mesure
métal
météore
méthode
métier
meuble
miauler
microbe
miette
mignon
migrer
milieu
million
mimique
mince
minéral
minimal
minorer
minute
miracle
miroiter
missile
mixte
mobile
moderne
moelleux
mondial
moniteur
monnaie
monotone
monstre
montagne
monument
moqueur
morceau
morsure
mortier
moteur
motif
mouche
moufle
moulin
mousson
mouton
mouvant
multiple
munition
muraille
murène
murmure
muscle
muséum
musicien
mutation
muter
mutuel
myriade
myrtille
mystère
mythique
nageur
nappe
narquois
narrer
natation
nation
nature
naufrage
nautique
navire
nébuleux
nectar
néfaste
négation
négliger
négocier
neige
nerveux
nettoyer
neurone
neutron
neveu
niche
nickel
nitrate
niveau
noble
nocif
nocturne
noirceur
noisette
nomade
nombreux
nommer
normatif
notable
notifier
notoire
nourrir
nouveau
novateur
novembre
novice
nuage
nuancer
nuire
nuisible
numéro
nuptial
nuque
nutritif
obéir
objectif
obliger
obscur
observer
obstacle
obtenir
obturer
occasion
occuper
océan
octobre
octroyer
octupler
oculaire
odeur
odorant
offenser
officier
offrir
ogive
oiseau
oisillon
olfactif
olivier
ombrage
omettre
onctueux
onduler
onéreux
onirique
opale
opaque
opérer
opinion
opportun
opprimer
opter
optique
orageux
orange
orbite
ordonner
oreille
organe
orgueil
orifice
ornement
orque
ortie
osciller
osmose
ossature
otarie
ouragan
ourson
outil
outrager
ouvrage
ovation
oxyde
oxygène
ozone
paisible
palace
palmarès
palourde
palper
panache
panda
pangolin
paniquer
panneau
panorama
pantalon
papaye
papier
papoter
papyrus
paradoxe
parcelle
paresse
parfumer
parler
parole
parrain
parsemer
partager
parure
parvenir
passion
pastèque
paternel
patience
patron
pavillon
pavoiser
payer
paysage
peigne
peintre
pelage
pélican
pelle
pelouse
peluche
pendule
pénétrer
pénible
pensif
pénurie
pépite
péplum
perdrix
perforer
période
permuter
perplexe
persil
perte
peser
pétale
petit
pétrir
peuple
pharaon
phobie
phoque
photon
phrase
physique
piano
pictural
pièce
pierre
pieuvre
pilote
pinceau
pipette
piquer
pirogue
piscine
piston
pivoter
pixel
pizza
placard
plafond
plaisir
planer
plaque
plastron
plateau
pleurer
plexus
pliage
plomb
plonger
pluie
plumage
pochette
poésie
poète
pointe
poirier
poisson
poivre
polaire
policier
pollen
polygone
pommade
pompier
ponctuel
pondérer
poney
portique
position
posséder
posture
potager
poteau
potion
pouce
poulain
poumon
pourpre
poussin
pouvoir
prairie
pratique
précieux
prédire
préfixe
prélude
prénom
présence
prétexte
prévoir
primitif
prince
prison
priver
problème
procéder
prodige
profond
progrès
proie
projeter
prologue
promener
propre
prospère
protéger
prouesse
proverbe
prudence
pruneau
psychose
public
puceron
puiser
pulpe
pulsar
punaise
punitif
pupitre
purifier
puzzle
pyramide
quasar
querelle
question
quiétude
quitter
quotient
racine
raconter
radieux
ragondin
raideur
raisin
ralentir
rallonge
ramasser
rapide
rasage
ratisser
ravager
ravin
rayonner
réactif
réagir
réaliser
réanimer
recevoir
réciter
réclamer
récolter
recruter
reculer
recycler
rédiger
redouter
refaire
réflexe
réformer
refrain
refuge
régalien
région
réglage
régulier
réitérer
rejeter
rejouer
relatif
relever
relief
remarque
remède
remise
remonter
remplir
remuer
renard
renfort
renifler
renoncer
rentrer
renvoi
replier
reporter
reprise
reptile
requin
réserve
résineux
résoudre
respect
rester
résultat
rétablir
retenir
réticule
retomber
retracer
réunion
réussir
revanche
revivre
révolte
révulsif
richesse
rideau
rieur
rigide
rigoler
rincer
riposter
risible
risque
rituel
rival
rivière
rocheux
romance
rompre
ronce
rondin
roseau
rosier
rotatif
rotor
rotule
rouge
rouille
rouleau
routine
royaume
ruban
rubis
ruche
ruelle
rugueux
ruiner
ruisseau
ruser
rustique
rythme
sabler
saboter
sabre
sacoche
safari
sagesse
saisir
salade
salive
salon
saluer
samedi
sanction
sanglier
sarcasme
sardine
saturer
saugrenu
saumon
sauter
sauvage
savant
savonner
scalpel
scandale
scélérat
scénario
sceptre
schéma
science
scinder
score
scrutin
sculpter
séance
sécable
sécher
secouer
sécréter
sédatif
séduire
seigneur
séjour
sélectif
semaine
sembler
semence
séminal
sénateur
sensible
sentence
séparer
séquence
serein
sergent
sérieux
serrure
sérum
service
sésame
sévir
sevrage
sextuple
sidéral
siècle
siéger
siffler
sigle
signal
silence
silicium
simple
sincère
sinistre
siphon
sirop
sismique
situer
skier
social
socle
sodium
soigneux
soldat
soleil
solitude
soluble
sombre
sommeil
somnoler
sonde
songeur
sonnette
sonore
sorcier
sortir
sosie
sottise
soucieux
soudure
souffle
soulever
soupape
source
soutirer
souvenir
spacieux
spatial
spécial
sphère
spiral
stable
station
sternum
stimulus
stipuler
strict
studieux
stupeur
styliste
sublime
substrat
subtil
subvenir
succès
sucre
suffixe
suggérer
suiveur
sulfate
superbe
supplier
surface
suricate
surmener
surprise
sursaut
survie
suspect
syllabe
symbole
symétrie
synapse
syntaxe
système
tabac
tablier
tactile
tailler
talent
talisman
talonner
tambour
tamiser
tangible
tapis
taquiner
tarder
tarif
tartine
tasse
tatami
tatouage
taupe
taureau
taxer
témoin
temporel
tenaille
tendre
teneur
tenir
tension
terminer
terne
terrible
tétine
texte
thème
théorie
thérapie
thorax
tibia
tiède
timide
tirelire
tiroir
tissu
titane
titre
tituber
toboggan
tolérant
tomate
tonique
tonneau
toponyme
torche
tordre
tornade
torpille
torrent
torse
tortue
totem
toucher
tournage
tousser
toxine
traction
trafic
tragique
trahir
train
trancher
travail
trèfle
tremper
trésor
treuil
triage
tribunal
tricoter
trilogie
triomphe
tripler
triturer
trivial
trombone
tronc
tropical
troupeau
tuile
tulipe
tumulte
tunnel
turbine
tuteur
tutoyer
tuyau
tympan
typhon
typique
tyran
ubuesque
ultime
ultrason
unanime
unifier
union
unique
unitaire
univers
uranium
urbain
urticant
usage
usine
usuel
usure
utile
utopie
vacarme
vaccin
vagabond
vague
vaillant
vaincre
vaisseau
valable
valise
vallon
valve
vampire
vanille
vapeur
varier
vaseux
vassal
vaste
vecteur
vedette
végétal
véhicule
veinard
véloce
vendredi
vénérer
venger
venimeux
ventouse
verdure
vérin
vernir
verrou
verser
vertu
veston
vétéran
vétuste
vexant
vexer
viaduc
viande
victoire
vidange
vidéo
vignette
vigueur
vilain
village
vinaigre
violon
vipère
virement
virtuose
virus
visage
viseur
vision
visqueux
visuel
vital
vitesse
viticole
vitrine
vivace
vivipare
vocation
voguer
voile
voisin
voiture
volaille
volcan
voltiger
volume
vorace
vortex
voter
vouloir
voyage
voyelle
wagon
xénon
yacht
zèbre
zénith
zeste
zoologie
"""
words_list = [x for x in s.split("\n") if x is not ""]